@@ -128,7 +128,7 @@ static int sink_callback(pa_context *c, struct global *g, struct sink_data *d)
 	i.owner_module = g->id;
 	i.volume.channels = i.sample_spec.channels;
 	for (n = 0; n < i.volume.channels; n++)
-		i.volume.values[n] = g->node_info.volume * g->node_info.channel_volumes[n] * PA_VOLUME_NORM;
+		i.volume.values[n] = pa_sw_volume_from_linear(g->node_info.volume * g->node_info.channel_volumes[n]);;
 	i.mute = g->node_info.mute;
 	i.monitor_source = g->node_info.monitor;
 	snprintf(monitor_name, sizeof(monitor_name)-1, "%s.monitor", i.name);
@@ -145,7 +145,7 @@ static int sink_callback(pa_context *c, struct global *g, struct sink_data *d)
 		  i.flags |= PA_SINK_HW_MUTE_CTRL;
 	i.proplist = pa_proplist_new_dict(info->props);
 	i.configured_latency = 0;
-	i.base_volume = g->node_info.base_volume * PA_VOLUME_NORM;
+	i.base_volume = pa_sw_volume_from_linear(g->node_info.base_volume);
 	i.n_volume_steps = g->node_info.volume_step * (PA_VOLUME_NORM+1);
 	i.state = node_state_to_sink(info->state);
 	i.card = g->node_info.device_id;
@@ -167,6 +167,7 @@ static int sink_callback(pa_context *c, struct global *g, struct sink_data *d)
 			if (!has_device(&cg->card_info.port_devices[n], g->node_info.profile_device_id))
 				continue;
 
+			spa_zero(spi[j]);
 			i.ports[j] = &spi[j];
 			spi[j].name = ci->ports[n]->name;
 			spi[j].description = ci->ports[n]->description;
@@ -182,6 +183,8 @@ static int sink_callback(pa_context *c, struct global *g, struct sink_data *d)
 		else
 			i.ports[j] = NULL;
 	}
+	if (i.active_port == NULL && i.n_ports > 0)
+		i.active_port = i.ports[0];
 	i.n_formats = pw_array_get_len(&g->node_info.formats, pa_format_info *);
 	i.formats = g->node_info.formats.data;
 	d->cb(c, &i, 0, d->userdata);
@@ -313,7 +316,7 @@ static int set_stream_volume(pa_context *c, pa_stream *s, const pa_cvolume *volu
 
 	if (volume) {
 		for (i = 0; i < volume->channels; i++)
-			channel_volumes[i] = volume->values[i] / (float) PA_VOLUME_NORM;
+			channel_volumes[i] = pa_sw_volume_to_linear(volume->values[i]);;
 		vols = channel_volumes;
 		n_channel_volumes = volume->channels;
 	} else {
@@ -343,7 +346,7 @@ static int set_node_volume(pa_context *c, struct global *g, const pa_cvolume *vo
 
 	if (volume) {
 		for (i = 0; i < volume->channels; i++)
-			channel_volumes[i] = volume->values[i] / (float) PA_VOLUME_NORM;
+			channel_volumes[i] = pa_sw_volume_to_linear(volume->values[i]);
 		vols = channel_volumes;
 		n_channel_volumes = volume->channels;
 
@@ -361,10 +364,12 @@ static int set_node_volume(pa_context *c, struct global *g, const pa_cvolume *vo
 			return 0;
 	}
 	g->node_info.mute = mute;
+	g->changed++;
 
 	if (!SPA_FLAG_IS_SET(g->permissions, PW_PERM_W | PW_PERM_X))
 		return PA_ERR_ACCESS;
 
+	pw_log_debug("node %p: id:%u", g, g->id);
 	pw_node_set_param((struct pw_node*)g->proxy,
 		SPA_PARAM_Props, 0,
 		spa_pod_builder_add_object(&b,
@@ -390,7 +395,7 @@ static int set_device_volume(pa_context *c, struct global *g, struct global *cg,
 
 	if (volume) {
 		for (i = 0; i < volume->channels; i++)
-			channel_volumes[i] = volume->values[i] / (float) PA_VOLUME_NORM;
+			channel_volumes[i] = pa_sw_volume_to_linear(volume->values[i]);
 		vols = channel_volumes;
 		n_channel_volumes = volume->channels;
 
@@ -408,6 +413,7 @@ static int set_device_volume(pa_context *c, struct global *g, struct global *cg,
 			return 0;
 	}
 	g->node_info.mute = mute;
+	g->changed++;
 
 	if (!SPA_FLAG_IS_SET(cg->permissions, PW_PERM_W | PW_PERM_X))
 		return PA_ERR_ACCESS;
@@ -428,6 +434,7 @@ static int set_device_volume(pa_context *c, struct global *g, struct global *cg,
 								vols));
 	param = spa_pod_builder_pop(&b, &f[0]);
 
+	pw_log_debug("device %p: id:%u", cg, cg->id);
 	pw_device_set_param((struct pw_node*)cg->proxy,
 		SPA_PARAM_Route, 0, param);
 
@@ -444,7 +451,7 @@ static int set_volume(pa_context *c, struct global *g, const pa_cvolume *volume,
 	card_id = g->node_info.device_id;
 	device_id = g->node_info.profile_device_id;
 
-	pw_log_info("card:%u global:%u flags:%08x", card_id, g->id, g->node_info.flags);
+	pw_log_debug("card:%u global:%u flags:%08x", card_id, g->id, g->node_info.flags);
 
 	if (SPA_FLAG_IS_SET(g->node_info.flags, NODE_FLAG_DEVICE_VOLUME | NODE_FLAG_DEVICE_MUTE) &&
 	    (cg = pa_context_find_global(c, card_id)) != NULL) {
@@ -612,6 +619,8 @@ pa_operation* pa_context_suspend_sink_by_name(pa_context *c, const char *sink_na
 	pa_operation *o;
 	struct success_ack *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: name:%s suspend:%d", c, sink_name, suspend);
 	o = pa_operation_new(c, NULL, on_success, sizeof(struct success_ack));
 	d = o->userdata;
@@ -629,6 +638,8 @@ pa_operation* pa_context_suspend_sink_by_index(pa_context *c, uint32_t idx, int
 	pa_operation *o;
 	struct success_ack *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: index:%u suspend:%d", c, idx, suspend);
 	o = pa_operation_new(c, NULL, on_success, sizeof(struct success_ack));
 	d = o->userdata;
@@ -827,7 +838,8 @@ static int source_callback(pa_context *c, struct global *g, struct source_data *
 		if ((str = spa_dict_lookup(info->props, PW_KEY_NODE_NAME)))
 			pa_proplist_setf(i.proplist, PW_KEY_NODE_NAME, "%s.monitor", str);
 		if ((str = spa_dict_lookup(info->props, PW_KEY_NODE_DESCRIPTION)))
-			pa_proplist_setf(i.proplist, PW_KEY_NODE_DESCRIPTION, "Monitor or %s", str);
+			pa_proplist_setf(i.proplist, PW_KEY_NODE_DESCRIPTION, "Monitor of %s", str);
+		pa_proplist_setf(i.proplist, PW_KEY_DEVICE_CLASS, "monitor");
 	}
 
 	if ((str = pa_proplist_gets(i.proplist, PW_KEY_NODE_NAME)))
@@ -835,7 +847,8 @@ static int source_callback(pa_context *c, struct global *g, struct source_data *
 	else
 		i.name = "unknown";
 
-	i.index = g->id;
+	pw_log_debug("source %d %s monitor:%d", g->id, i.name, monitor);
+
 	if ((str = pa_proplist_gets(i.proplist, PW_KEY_NODE_DESCRIPTION)))
 		i.description = str;
 	else
@@ -853,13 +866,14 @@ static int source_callback(pa_context *c, struct global *g, struct source_data *
 	i.owner_module = g->id;
 	i.volume.channels = i.sample_spec.channels;
 	for (n = 0; n < i.volume.channels; n++)
-		i.volume.values[n] = g->node_info.volume * g->node_info.channel_volumes[n] * PA_VOLUME_NORM;
+		i.volume.values[n] = pa_sw_volume_from_linear(g->node_info.volume * g->node_info.channel_volumes[n]);;
 	i.mute = g->node_info.mute;
 	if (monitor) {
+		i.index = g->node_info.monitor;
 		i.monitor_of_sink = g->id;
 		i.monitor_of_sink_name = pa_context_find_global_name(c, g->id);
-		i.index = g->node_info.monitor;
 	} else {
+		i.index = g->id;
 		i.monitor_of_sink = PA_INVALID_INDEX;
 		i.monitor_of_sink_name = NULL;
 		if (info->props && (str = spa_dict_lookup(info->props, PW_KEY_DEVICE_API)))
@@ -873,7 +887,7 @@ static int source_callback(pa_context *c, struct global *g, struct source_data *
 	i.driver = "PipeWire";
 	i.flags = flags;
 	i.configured_latency = 0;
-	i.base_volume = g->node_info.base_volume * PA_VOLUME_NORM;
+	i.base_volume = pa_sw_volume_from_linear(g->node_info.base_volume);
 	i.n_volume_steps = g->node_info.volume_step * (PA_VOLUME_NORM+1);
 	i.state = node_state_to_source(info->state);
 	i.card = g->node_info.device_id;
@@ -894,6 +908,8 @@ static int source_callback(pa_context *c, struct global *g, struct source_data *
 				continue;
 			if (!has_device(&cg->card_info.port_devices[n], g->node_info.profile_device_id))
 				continue;
+
+			spa_zero(spi[j]);
 			i.ports[j] = &spi[j];
 			spi[j].name = ci->ports[n]->name;
 			spi[j].description = ci->ports[n]->description;
@@ -909,6 +925,8 @@ static int source_callback(pa_context *c, struct global *g, struct source_data *
 		else
 			i.ports[j] = NULL;
 	}
+	if (i.active_port == NULL && i.n_ports > 0)
+		i.active_port = i.ports[0];
 	i.n_formats = pw_array_get_len(&g->node_info.formats, pa_format_info *);
 	i.formats = g->node_info.formats.data;
 	d->cb(c, &i, 0, d->userdata);
@@ -979,6 +997,7 @@ pa_operation* pa_context_get_source_info_by_index(pa_context *c, uint32_t idx, p
 	pa_assert(c->refcount >= 1);
 	pa_assert(cb);
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
 	PA_CHECK_VALIDITY_RETURN_NULL(c, idx != PA_INVALID_INDEX, PA_ERR_INVALID);
 
 	pw_log_debug("context %p: index %d", c, idx);
@@ -1144,6 +1163,8 @@ pa_operation* pa_context_suspend_source_by_name(pa_context *c, const char *sourc
 	pa_operation *o;
 	struct success_ack *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: name:%s", c, source_name);
 
 	o = pa_operation_new(c, NULL, on_success, sizeof(struct success_ack));
@@ -1162,6 +1183,8 @@ pa_operation* pa_context_suspend_source_by_index(pa_context *c, uint32_t idx, in
 	pa_operation *o;
 	struct success_ack *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: index:%u", c, idx);
 	o = pa_operation_new(c, NULL, on_success, sizeof(struct success_ack));
 	d = o->userdata;
@@ -1295,6 +1318,8 @@ pa_operation* pa_context_get_server_info(pa_context *c, pa_server_info_cb_t cb,
 	pa_assert(c->refcount >= 1);
 	pa_assert(cb);
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p", c);
 	o = pa_operation_new(c, NULL, server_info, sizeof(struct server_data));
 	d = o->userdata;
@@ -1349,6 +1374,7 @@ pa_operation* pa_context_get_module_info(pa_context *c, uint32_t idx, pa_module_
 	pa_assert(c->refcount >= 1);
 	pa_assert(cb);
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
 	PA_CHECK_VALIDITY_RETURN_NULL(c, idx != PA_INVALID_INDEX, PA_ERR_INVALID);
 
 	pw_log_debug("context %p index:%u", c, idx);
@@ -1478,9 +1504,9 @@ static void module_proxy_error(void *data, int seq, int res, const char *message
 	on_load_module(o, d);
 }
 
-static void on_null_sink_module(pa_operation *o, void *userdata)
+static int load_null_sink_module(pa_operation *o)
 {
-	struct load_module *d = userdata;
+	struct load_module *d = o->userdata;
 	pa_context *c = o->context;
 	static const struct pw_proxy_events proxy_events = {
 		.bound = module_proxy_bound,
@@ -1488,15 +1514,18 @@ static void on_null_sink_module(pa_operation *o, void *userdata)
 	};
 
 	if (d->proxy != NULL)
-		return;
+		return -EBUSY;
 
 	d->proxy = pw_core_create_object(c->core,
                                 "adapter",
                                 PW_TYPE_INTERFACE_Node,
                                 PW_VERSION_NODE,
                                 d->props ? &d->props->dict : NULL, 0);
+	if (d->proxy == NULL)
+		return -errno;
 
 	pw_proxy_add_listener(d->proxy, &d->listener, &proxy_events, o);
+	return 0;
 }
 
 static void add_props(struct pw_properties *props, const char *str)
@@ -1536,15 +1565,23 @@ pa_operation* pa_context_load_module(pa_context *c, const char*name, const char
 	struct load_module *d;
 	int error = PA_ERR_NOTIMPLEMENTED;;
 	struct pw_properties *props = NULL;
-	pa_operation_cb_t op_cb = on_load_module;
 	const char *str;
+	bool sync = true;
 
 	pa_assert(c);
 	pa_assert(c->refcount >= 1);
 	pa_assert(name != NULL);
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: name:%s arg:%s", c, name, argument);
 
+	o = pa_operation_new(c, NULL, on_load_module, sizeof(struct load_module));
+	d = o->userdata;
+	d->cb = cb;
+	d->userdata = userdata;
+	d->idx = PA_INVALID_INDEX;
+
 	if (strcmp(name, "module-null-sink") == 0) {
 		props = pw_properties_new_string(argument);
 		if (props == NULL) {
@@ -1567,18 +1604,14 @@ pa_operation* pa_context_load_module(pa_context *c, const char*name, const char
 		}
 		pw_properties_set(props, "factory.name", "support.null-audio-sink");
 
-		error = 0;
-		op_cb = on_null_sink_module;
+		d->props = props;
+		error = load_null_sink_module(o);
+		sync = error < 0;
 	}
 done:
-	o = pa_operation_new(c, NULL, op_cb, sizeof(struct load_module));
-	d = o->userdata;
-	d->cb = cb;
 	d->error = error;
-	d->userdata = userdata;
-	d->idx = PA_INVALID_INDEX;
-	d->props = props;
-	pa_operation_sync(o);
+	if (sync)
+		pa_operation_sync(o);
 
 	return o;
 }
@@ -1591,6 +1624,8 @@ pa_operation* pa_context_unload_module(pa_context *c, uint32_t idx, pa_context_s
 	struct module_info *m;
 	int error;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: %u", c, idx);
 	if ((m = find_module(c, idx)) != NULL) {
 		pw_proxy_destroy(m->proxy);
@@ -1653,6 +1688,7 @@ pa_operation* pa_context_get_client_info(pa_context *c, uint32_t idx, pa_client_
 	pa_assert(c->refcount >= 1);
 	pa_assert(cb);
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
 	PA_CHECK_VALIDITY_RETURN_NULL(c, idx != PA_INVALID_INDEX, PA_ERR_INVALID);
 
 	pw_log_debug("context %p: index:%u", c, idx);
@@ -1738,6 +1774,7 @@ pa_operation* pa_context_kill_client(pa_context *c, uint32_t idx, pa_context_suc
 	pa_operation *o;
 	struct kill_client *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
 	PA_CHECK_VALIDITY_RETURN_NULL(c, idx != PA_INVALID_INDEX, PA_ERR_INVALID);
 
 	pw_log_debug("context %p: index:%u", c, idx);
@@ -1993,6 +2030,8 @@ pa_operation* pa_context_set_port_latency_offset(pa_context *c, const char *card
 	pa_operation *o;
 	struct success_ack *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("%p: card_name:%s port_name:%s offset:%"PRIi64, c, card_name, port_name, offset);
 	o = pa_operation_new(c, NULL, on_success, sizeof(struct success_ack));
 	d = o->userdata;
@@ -2073,7 +2112,7 @@ static int sink_input_callback(pa_context *c, struct sink_input_data *d, struct
 	pa_cvolume_init(&i.volume);
 	i.volume.channels = i.sample_spec.channels;
 	for (n = 0; n < i.volume.channels; n++)
-		i.volume.values[n] = g->node_info.volume * g->node_info.channel_volumes[n] * PA_VOLUME_NORM;
+		i.volume.values[n] = pa_sw_volume_from_linear(g->node_info.volume * g->node_info.channel_volumes[n]);
 
 	i.mute = g->node_info.mute;
 	i.buffer_usec = 0;
@@ -2127,6 +2166,7 @@ pa_operation* pa_context_get_sink_input_info(pa_context *c, uint32_t idx, pa_sin
 	pa_assert(c->refcount >= 1);
 	pa_assert(cb);
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
 	PA_CHECK_VALIDITY_RETURN_NULL(c, idx != PA_INVALID_INDEX, PA_ERR_INVALID);
 
 	pw_log_debug("context %p: info for %d", c, idx);
@@ -2238,6 +2278,8 @@ pa_operation* pa_context_move_sink_input_by_name(pa_context *c, uint32_t idx, co
 	pa_operation *o;
 	struct target_node *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("%p: index:%u name:%s", c, idx, sink_name);
 	o = pa_operation_new(c, NULL, do_target_node, sizeof(struct target_node));
 	d = o->userdata;
@@ -2259,6 +2301,8 @@ pa_operation* pa_context_move_sink_input_by_index(pa_context *c, uint32_t idx, u
 	pa_operation *o;
 	struct target_node *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("%p: index:%u sink_index:%u", c, idx, sink_idx);
 	o = pa_operation_new(c, NULL, do_target_node, sizeof(struct target_node));
 	d = o->userdata;
@@ -2322,6 +2366,8 @@ pa_operation* pa_context_set_sink_input_volume(pa_context *c, uint32_t idx, cons
 	pa_operation *o;
 	struct stream_volume *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: index %d", c, idx);
 	o = pa_operation_new(c, NULL, do_stream_volume_mute, sizeof(struct stream_volume));
 	d = o->userdata;
@@ -2342,6 +2388,8 @@ pa_operation* pa_context_set_sink_input_mute(pa_context *c, uint32_t idx, int mu
 	pa_operation *o;
 	struct stream_volume *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: index %d", c, idx);
 	o = pa_operation_new(c, NULL, do_stream_volume_mute, sizeof(struct stream_volume));
 	d = o->userdata;
@@ -2395,6 +2443,8 @@ pa_operation* pa_context_kill_sink_input(pa_context *c, uint32_t idx, pa_context
 	pa_operation *o;
 	struct kill_stream *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: index %d", c, idx);
 	o = pa_operation_new(c, NULL, do_kill_stream, sizeof(struct kill_stream));
 	d = o->userdata;
@@ -2476,7 +2526,7 @@ static int source_output_callback(struct source_output_data *d, pa_context *c, s
 	pa_cvolume_init(&i.volume);
 	i.volume.channels = i.sample_spec.channels;
 	for (n = 0; n < i.volume.channels; n++)
-		i.volume.values[n] = g->node_info.volume * g->node_info.channel_volumes[n] * PA_VOLUME_NORM;
+		i.volume.values[n] = pa_sw_volume_from_linear(g->node_info.volume * g->node_info.channel_volumes[n]);
 
 	i.mute = g->node_info.mute;
 	i.buffer_usec = 0;
@@ -2528,6 +2578,7 @@ pa_operation* pa_context_get_source_output_info(pa_context *c, uint32_t idx, pa_
 	pa_assert(c->refcount >= 1);
 	pa_assert(cb);
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
 	PA_CHECK_VALIDITY_RETURN_NULL(c, idx != PA_INVALID_INDEX, PA_ERR_INVALID);
 
 	pw_log_debug("%p: index:%u", c, idx);
@@ -2584,6 +2635,8 @@ pa_operation* pa_context_move_source_output_by_name(pa_context *c, uint32_t idx,
 	pa_operation *o;
 	struct target_node *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("%p index:%u name:%s", c, idx, source_name);
 	o = pa_operation_new(c, NULL, do_target_node, sizeof(struct target_node));
 	d = o->userdata;
@@ -2605,6 +2658,8 @@ pa_operation* pa_context_move_source_output_by_index(pa_context *c, uint32_t idx
 	pa_operation *o;
 	struct target_node *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("%p index:%u source_index:%u", c, idx, source_idx);
 	o = pa_operation_new(c, NULL, do_target_node, sizeof(struct target_node));
 	d = o->userdata;
@@ -2626,6 +2681,8 @@ pa_operation* pa_context_set_source_output_volume(pa_context *c, uint32_t idx, c
 	pa_operation *o;
 	struct stream_volume *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: index %d", c, idx);
 	o = pa_operation_new(c, NULL, do_stream_volume_mute, sizeof(struct stream_volume));
 	d = o->userdata;
@@ -2646,6 +2703,8 @@ pa_operation* pa_context_set_source_output_mute(pa_context *c, uint32_t idx, int
 	pa_operation *o;
 	struct stream_volume *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: index %d", c, idx);
 	o = pa_operation_new(c, NULL, do_stream_volume_mute, sizeof(struct stream_volume));
 	d = o->userdata;
@@ -2665,6 +2724,8 @@ pa_operation* pa_context_kill_source_output(pa_context *c, uint32_t idx, pa_cont
 	pa_operation *o;
 	struct kill_stream *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("context %p: index %d", c, idx);
 	o = pa_operation_new(c, NULL, do_kill_stream, sizeof(struct kill_stream));
 	d = o->userdata;
@@ -2702,6 +2763,8 @@ pa_operation* pa_context_stat(pa_context *c, pa_stat_info_cb_t cb, void *userdat
 	pa_operation *o;
 	struct stat_ack *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("%p", c);
 	o = pa_operation_new(c, NULL, on_stat_info, sizeof(struct stat_ack));
 	d = o->userdata;
@@ -2736,6 +2799,8 @@ pa_operation* pa_context_get_sample_info_by_name(pa_context *c, const char *name
 	pa_operation *o;
 	struct sample_info *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("%p nane:%s", c, name);
 	o = pa_operation_new(c, NULL, on_sample_info, sizeof(struct sample_info));
 	d = o->userdata;
@@ -2753,6 +2818,8 @@ pa_operation* pa_context_get_sample_info_by_index(pa_context *c, uint32_t idx, p
 	pa_operation *o;
 	struct sample_info *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("%p index:%u", c, idx);
 	o = pa_operation_new(c, NULL, on_sample_info, sizeof(struct sample_info));
 	d = o->userdata;
@@ -2782,6 +2849,8 @@ pa_operation* pa_context_get_sample_info_list(pa_context *c, pa_sample_info_cb_t
 	pa_operation *o;
 	struct sample_info *d;
 
+	PA_CHECK_VALIDITY_RETURN_NULL(c, c->state == PA_CONTEXT_READY, PA_ERR_BADSTATE);
+
 	pw_log_debug("%p", c);
 	o = pa_operation_new(c, NULL, on_sample_info_list, sizeof(struct sample_info));
 	d = o->userdata;
