@@ -32,12 +32,44 @@ void *_acp_log_data;
 
 #define VOLUME_ACCURACY (PA_VOLUME_NORM/100)  /* don't require volume adjustments to be perfectly correct. don't necessarily extend granularity in software unless the differences get greater than this level */
 
-static void profile_free(void *data)
+const char *acp_available_str(enum acp_available status)
 {
+	switch (status) {
+	case ACP_AVAILABLE_UNKNOWN:
+		return "unknown";
+	case ACP_AVAILABLE_NO:
+		return "no";
+	case ACP_AVAILABLE_YES:
+		return "yes";
+	}
+	return "error";
+}
+
+const char *acp_direction_str(enum acp_direction direction)
+{
+	switch (direction) {
+	case ACP_DIRECTION_CAPTURE:
+		return "capture";
+	case ACP_DIRECTION_PLAYBACK:
+		return "playback";
+	}
+	return "error";
 }
 
 static void port_free(void *data)
 {
+	pa_device_port *dp = data;
+	pa_dynarray_clear(&dp->devices);
+	pa_dynarray_clear(&dp->prof);
+	pa_device_port_free(dp);
+}
+
+static void device_free(void *data)
+{
+	pa_alsa_device *dev = data;
+	pa_dynarray_clear(&dev->port_array);
+	pa_proplist_free(dev->proplist);
+	pa_hashmap_free(dev->ports);
 }
 
 static void init_device(pa_card *impl, pa_alsa_device *dev, pa_alsa_direction_t direction,
@@ -77,13 +109,11 @@ static void init_device(pa_card *impl, pa_alsa_device *dev, pa_alsa_direction_t
 	pa_proplist_setf(dev->proplist, "card.profile.device", "%u", index);
 	pa_proplist_as_dict(dev->proplist, &dev->device.props);
 
-	dev->ports = pa_hashmap_new_full(pa_idxset_string_hash_func,
-			pa_idxset_string_compare_func, NULL,
-			(pa_free_cb_t) port_free);
+	dev->ports = pa_hashmap_new(pa_idxset_string_hash_func,
+			pa_idxset_string_compare_func);
 	if (m->ucm_context.ucm)
 		dev->ucm_context = &m->ucm_context;
 	pa_dynarray_init(&dev->port_array, NULL);
-
 }
 
 static int compare_profile(const void *a, const void *b)
@@ -100,6 +130,17 @@ static int compare_profile(const void *a, const void *b)
 	return p2->profile.priority - p1->profile.priority;
 }
 
+static void profile_free(void *data)
+{
+	pa_alsa_profile *ap = data;
+	pa_dynarray_clear(&ap->out.devices);
+	if (ap->profile.flags & ACP_PROFILE_OFF) {
+		free(ap->name);
+		free(ap->description);
+		free(ap);
+	}
+}
+
 static void add_profiles(pa_card *impl)
 {
 	pa_alsa_profile *ap;
@@ -111,13 +152,13 @@ static void add_profiles(pa_card *impl)
 	uint32_t idx;
 
 	n_devices = 0;
-	pa_dynarray_init(&impl->out.devices, NULL);
+	pa_dynarray_init(&impl->out.devices, device_free);
 
 	ap = pa_xnew0(pa_alsa_profile, 1);
 	ap->profile.name = ap->name = pa_xstrdup("off");
 	ap->profile.description = ap->description = pa_xstrdup(_("Off"));
 	ap->profile.available = ACP_AVAILABLE_YES;
-	ap->is_off = true;
+	ap->profile.flags = ACP_PROFILE_OFF;
 	pa_hashmap_put(impl->profiles, ap->name, ap);
 
 	PA_HASHMAP_FOREACH(ap, impl->profile_set->profiles, state) {
@@ -264,19 +305,18 @@ static pa_available_t calc_port_state(pa_device_port *p, pa_card *impl)
 }
 
 static void profile_set_available(pa_card *impl, uint32_t index,
-		enum acp_available status)
+		enum acp_available status, bool emit)
 {
 	struct acp_card_profile *p = impl->card.profiles[index];
 	enum acp_available old = p->available;
 
-	pa_log_debug("Setting profile %s to availability status %d",
-			p->name, status);
-	if (old == status)
-		return;
+	if (old != status)
+		pa_log_info("Profile %s available %s -> %s", p->name,
+				acp_available_str(old), acp_available_str(status));
 
 	p->available = status;
 
-	if (impl && impl->events && impl->events->profile_available)
+	if (emit && impl && impl->events && impl->events->profile_available)
 		impl->events->profile_available(impl->user_data, index,
 				old, status);
 }
@@ -411,6 +451,9 @@ static int report_jack_state(snd_mixer_elem_t *melem, unsigned int mask)
 		bool found_available_output_port = false;
 		enum acp_available available = ACP_AVAILABLE_UNKNOWN;
 
+		if (profile->profile.flags & ACP_PROFILE_OFF)
+			continue;
+
 		PA_HASHMAP_FOREACH(port, impl->ports, state2) {
 			if (!pa_hashmap_get(port->profiles, profile->profile.name))
 				continue;
@@ -421,7 +464,6 @@ static int report_jack_state(snd_mixer_elem_t *melem, unsigned int mask)
 					found_available_input_port = true;
 			} else {
 				has_output_port = true;
-
 				if (port->port.available != ACP_AVAILABLE_NO)
 					found_available_output_port = true;
 			}
@@ -444,11 +486,11 @@ static int report_jack_state(snd_mixer_elem_t *melem, unsigned int mask)
 		if (profile->profile.index == impl->card.active_profile_index)
 			active_available = available;
 		else
-			profile_set_available(impl, profile->profile.index, available);
+			profile_set_available(impl, profile->profile.index, available, false);
 	}
 
 	if (impl->card.active_profile_index != ACP_INVALID_INDEX)
-		profile_set_available(impl, impl->card.active_profile_index, active_available);
+		profile_set_available(impl, impl->card.active_profile_index, active_available, true);
 
 	return 0;
 }
@@ -565,9 +607,10 @@ static int hdmi_eld_changed(snd_mixer_elem_t *melem, unsigned int mask)
 		changed |= (old_monitor_name == NULL) || (strcmp(old_monitor_name, eld.monitor_name) != 0);
 		pa_proplist_sets(p->proplist, PA_PROP_DEVICE_PRODUCT_NAME, eld.monitor_name);
 	}
+	pa_proplist_as_dict(p->proplist, &p->port.props);
 
 	if (changed && mask != 0 && impl->events && impl->events->props_changed)
-		impl->events->props_changed(impl);
+		impl->events->props_changed(impl->user_data);
 	return 0;
 }
 
@@ -622,10 +665,11 @@ static void init_eld_ctls(pa_card *impl)
 uint32_t acp_card_find_best_profile_index(struct acp_card *card, const char *name)
 {
 	uint32_t i;
-	uint32_t best, best2, best3;
+	uint32_t best, best2, off;
 	struct acp_card_profile **profiles = card->profiles;
 
-	best = best2 = best3 = ACP_INVALID_INDEX;
+	best = best2 = ACP_INVALID_INDEX;
+	off = 0;
 
 	for (i = 0; i < card->n_profiles; i++) {
 		struct acp_card_profile *p = profiles[i];
@@ -635,23 +679,20 @@ uint32_t acp_card_find_best_profile_index(struct acp_card *card, const char *nam
 				best = i;
 			continue;
 		}
-		if (p->available == ACP_AVAILABLE_YES) {
+		if (p->flags & ACP_PROFILE_OFF) {
+			off = i;
+		} else if (p->available == ACP_AVAILABLE_YES) {
 			if (best == ACP_INVALID_INDEX || p->priority > profiles[best]->priority)
 				best = i;
 		} else if (p->available != ACP_AVAILABLE_NO) {
 			if (best2 == ACP_INVALID_INDEX || p->priority > profiles[best2]->priority)
 				best2 = i;
-		} else {
-			if (best3 == ACP_INVALID_INDEX || p->priority > profiles[best3]->priority)
-				best3 = i;
 		}
 	}
 	if (best == ACP_INVALID_INDEX)
 		best = best2;
 	if (best == ACP_INVALID_INDEX)
-		best = best3;
-	if (best == ACP_INVALID_INDEX)
-		best = 0;
+		best = off;
 	return best;
 }
 
@@ -781,12 +822,19 @@ static void set_volume(pa_alsa_device *dev, const pa_cvolume *v)
 				pa_cvolume_max(&new_soft_volume),
 				pa_yes_no(accurate_enough));
 
-		if (!accurate_enough && impl->events && impl->events->set_soft_volume) {
-			uint32_t i, n_volumes = new_soft_volume.channels;
-			float volumes[n_volumes];
-			for (i = 0; i < n_volumes; i++)
-				volumes[i] = ((float)new_soft_volume.values[i]) / PA_VOLUME_NORM;
-			impl->events->set_soft_volume(impl->user_data, &dev->device, volumes, n_volumes);
+		if (accurate_enough)
+			pa_cvolume_reset(&new_soft_volume, new_soft_volume.channels);
+
+		if (!pa_cvolume_equal(&dev->soft_volume, &new_soft_volume)) {
+			dev->soft_volume = new_soft_volume;
+
+			if (impl->events && impl->events->set_soft_volume) {
+				uint32_t i, n_volumes = new_soft_volume.channels;
+				float volumes[n_volumes];
+				for (i = 0; i < n_volumes; i++)
+					volumes[i] = pa_sw_volume_to_linear(new_soft_volume.values[i]);
+				impl->events->set_soft_volume(impl->user_data, &dev->device, volumes, n_volumes);
+			}
 		}
 
 	} else {
@@ -868,7 +916,7 @@ static void mixer_volume_init(pa_alsa_device *dev)
 		pa_log_info("Using hardware volume control. Hardware dB scale %s.",
 				dev->mixer_path->has_dB ? "supported" : "not supported");
 	}
-	dev->device.base_volume = (float)dev->base_volume / PA_VOLUME_NORM;
+	dev->device.base_volume = pa_sw_volume_to_linear(dev->base_volume);;
 	dev->device.volume_step = 1.0f / dev->n_volume_steps;
 
 	if (!dev->mixer_path || !dev->mixer_path->has_mute) {
@@ -994,8 +1042,10 @@ static int device_enable(pa_card *impl, pa_alsa_mapping *mapping, pa_alsa_device
 	find_mixer(impl, dev, NULL, ignore_dB);
 
 	port_index = acp_device_find_best_port_index(&dev->device, NULL);
-
-	dev->active_port = (pa_device_port*)impl->card.ports[port_index];
+	if (port_index == ACP_INVALID_INDEX)
+		dev->active_port = NULL;
+	else
+		dev->active_port = (pa_device_port*)impl->card.ports[port_index];
 	if (dev->active_port)
 		dev->active_port->port.flags |= ACP_PORT_ACTIVE;
 
@@ -1025,6 +1075,11 @@ int acp_card_set_profile(struct acp_card *card, uint32_t new_index)
 	op = old_index != ACP_INVALID_INDEX ? (pa_alsa_profile*)profiles[old_index] : NULL;
 	np = (pa_alsa_profile*)profiles[new_index];
 
+	if (op == np)
+		return 0;
+
+	pa_log_info("activate profile: %s (%d)", np->profile.name, new_index);
+
 	if (op && op->output_mappings) {
 		PA_IDXSET_FOREACH(am, op->output_mappings, idx) {
 			if (np->output_mappings &&
@@ -1046,7 +1101,8 @@ int acp_card_set_profile(struct acp_card *card, uint32_t new_index)
 
 	/* if UCM is available for this card then update the verb */
 	if (impl->use_ucm) {
-		if ((res = pa_alsa_ucm_set_profile(&impl->ucm, impl, np->is_off ? NULL : np->profile.name,
+		if ((res = pa_alsa_ucm_set_profile(&impl->ucm, impl,
+		    np->profile.flags & ACP_PROFILE_OFF ? NULL : np->profile.name,
 		    op ? op->profile.name : NULL)) < 0) {
 			return res;
 		}
@@ -1068,8 +1124,6 @@ int acp_card_set_profile(struct acp_card *card, uint32_t new_index)
 	np->profile.flags |= ACP_PROFILE_ACTIVE;
 	impl->card.active_profile_index = new_index;
 
-	pa_log_info("active_profile: %s (%d)", np->profile.name, new_index);
-
 	if (impl->events && impl->events->profile_changed)
 		impl->events->profile_changed(impl->user_data, old_index,
 				new_index);
@@ -1094,6 +1148,7 @@ struct acp_card *acp_card_new(uint32_t index, const struct acp_dict *props)
 	char device_id[16];
 	bool ignore_dB = false;
 	uint32_t profile_index;
+	int res;
 
 	impl = calloc(1, sizeof(*impl));
 	if (impl == NULL)
@@ -1142,15 +1197,22 @@ struct acp_card *acp_card_new(uint32_t index, const struct acp_dict *props)
 
 	snd_config_update_free_global();
 
-	if (impl->use_ucm && !pa_alsa_ucm_query_profiles(&impl->ucm, card->index)) {
+	res = impl->use_ucm ? pa_alsa_ucm_query_profiles(&impl->ucm, card->index) : -1;
+	if (res == -PA_ALSA_ERR_UCM_LINKED) {
+		res = -ENOENT;
+		goto error;
+	}
+	if (res == 0) {
 		pa_log_info("Found UCM profiles");
 		impl->profile_set = pa_alsa_ucm_add_profile_set(&impl->ucm, &impl->ucm.default_channel_map);
 	} else {
 		impl->use_ucm = false;
 		impl->profile_set = pa_alsa_profile_set_new(profile_set, &impl->ucm.default_channel_map);
 	}
-	if (impl->profile_set == NULL)
-		return NULL;
+	if (impl->profile_set == NULL) {
+		res = -ENOTSUP;
+		goto error;
+	}
 
 	impl->profile_set->ignore_dB = ignore_dB;
 
@@ -1185,6 +1247,11 @@ struct acp_card *acp_card_new(uint32_t index, const struct acp_dict *props)
 	init_eld_ctls(impl);
 
 	return &impl->card;
+error:
+	pa_alsa_refcnt_dec();
+	free(impl);
+	errno = -res;
+	return NULL;
 }
 
 void acp_card_add_listener(struct acp_card *card,
@@ -1197,7 +1264,24 @@ void acp_card_add_listener(struct acp_card *card,
 
 void acp_card_destroy(struct acp_card *card)
 {
+	pa_card *impl = (pa_card *)card;
+	if (impl->profiles)
+		pa_hashmap_free(impl->profiles);
+	if (impl->ports)
+		pa_hashmap_free(impl->ports);
+	pa_dynarray_clear(&impl->out.devices);
+	pa_dynarray_clear(&impl->out.profiles);
+	pa_dynarray_clear(&impl->out.ports);
+	if (impl->ucm.mixers)
+		pa_hashmap_free(impl->ucm.mixers);
+	if (impl->jacks)
+		pa_hashmap_free(impl->jacks);
+	if (impl->profile_set)
+		pa_alsa_profile_set_free(impl->profile_set);
+	pa_alsa_ucm_free(&impl->ucm);
+	pa_proplist_free(impl->proplist);
 	pa_alsa_refcnt_dec();
+	free(impl);
 }
 
 int acp_card_poll_descriptors_count(struct acp_card *card)
@@ -1332,7 +1416,10 @@ uint32_t acp_device_find_best_port_index(struct acp_device *dev, const char *nam
 		best = best3;
 	if (best == ACP_INVALID_INDEX)
 		best = 0;
-	return ports[best]->index;
+	if (best < dev->n_ports)
+		return ports[best]->index;
+	else
+		return ACP_INVALID_INDEX;
 }
 
 int acp_device_set_port(struct acp_device *dev, uint32_t port_index)
@@ -1402,7 +1489,7 @@ int acp_device_set_volume(struct acp_device *dev, const float *volume, uint32_t
 
 	v.channels = d->mapping->channel_map.channels;
 	for (i = 0; i < v.channels; i++)
-		v.values[i] = volume[i % n_volume] * PA_VOLUME_NORM;
+		v.values[i] = pa_sw_volume_from_linear(volume[i % n_volume]);;
 
 	pa_log_info("Set %s volume: %d", d->set_volume ? "hardware" : "software", pa_cvolume_max(&v));
 	for (i = 0; i < v.channels; i++)
@@ -1412,6 +1499,7 @@ int acp_device_set_volume(struct acp_device *dev, const float *volume, uint32_t
 		d->set_volume(d, &v);
 	} else {
 		d->real_volume = v;
+		d->soft_volume = v;
 		if (impl->events && impl->events->set_soft_volume)
 			impl->events->set_soft_volume(impl->user_data, dev, volume, n_volume);
 	}
@@ -1430,7 +1518,7 @@ int acp_device_get_volume(struct acp_device *dev, float *volume, uint32_t n_volu
 	if (v.channels == 0)
 		return -EIO;
 	for (i = 0; i < n_volume; i++)
-		volume[i] = ((float)v.values[i % v.channels]) / PA_VOLUME_NORM;
+		volume[i] = pa_sw_volume_to_linear(v.values[i % v.channels]);
 	return 0;
 }
 
