@@ -662,7 +662,9 @@ static void card_profile_available(void *data, uint32_t index,
 	struct impl *this = data;
 	struct acp_card *card = this->card;
 	struct acp_card_profile *p = card->profiles[index];
-	spa_log_info(this->log, "card profile %s available %d", p->name, available);
+
+	spa_log_info(this->log, "card profile %s available %s -> %s", p->name,
+			acp_available_str(old), acp_available_str(available));
 
 	this->info.change_mask |= SPA_DEVICE_CHANGE_MASK_PARAMS;
 	this->params[IDX_EnumProfile].flags ^= SPA_PARAM_INFO_SERIAL;
@@ -696,7 +698,9 @@ static void card_port_available(void *data, uint32_t index,
 	struct impl *this = data;
 	struct acp_card *card = this->card;
 	struct acp_port *p = card->ports[index];
-	spa_log_info(this->log, "card port %s available %d", p->name, available);
+
+	spa_log_info(this->log, "card port %s available %s -> %s", p->name,
+			acp_available_str(old), acp_available_str(available));
 
 	this->info.change_mask |= SPA_DEVICE_CHANGE_MASK_PARAMS;
 	this->params[IDX_EnumRoute].flags ^= SPA_PARAM_INFO_SERIAL;
@@ -746,7 +750,7 @@ static void on_set_soft_volume(void *data, struct acp_device *dev,
 	struct spa_pod_builder b = { 0 };
 	struct spa_pod_frame f[1];
 
-	spa_log_debug(this->log, "device %s soft volume %f", dev->name, volume[0]);
+	spa_log_info(this->log, "device %s soft volume %f", dev->name, volume[0]);
 
 	spa_pod_builder_init(&b, buffer, sizeof(buffer));
 	spa_pod_builder_push_object(&b, &f[0],
@@ -772,7 +776,7 @@ static void on_set_soft_mute(void *data, struct acp_device *dev,
 	struct spa_pod_builder b = { 0 };
 	struct spa_pod_frame f[1];
 
-	spa_log_debug(this->log, "device %s soft mute %d", dev->name, mute);
+	spa_log_info(this->log, "device %s soft mute %d", dev->name, mute);
 
 	spa_pod_builder_init(&b, buffer, sizeof(buffer));
 	spa_pod_builder_push_object(&b, &f[0],
@@ -823,14 +827,18 @@ static SPA_PRINTF_FUNC(6,0) void impl_acp_log_func(void *data,
 		int level, const char *file, int line, const char *func,
 		const char *fmt, va_list arg)
 {
-	struct impl *this = data;
-	spa_log_logv(this->log, (enum spa_log_level)level, file, line, func, fmt, arg);
+	struct spa_log *log = data;
+	spa_log_logv(log, (enum spa_log_level)level, file, line, func, fmt, arg);
 }
 
 static int impl_clear(struct spa_handle *handle)
 {
 	struct impl *this = (struct impl *) handle;
 	remove_sources(this);
+	if (this->card) {
+		acp_card_destroy(this->card);
+		this->card = NULL;
+	}
 	return 0;
 }
 
@@ -869,7 +877,7 @@ impl_init(const struct spa_handle_factory *factory,
 		return -EINVAL;
 	}
 
-	acp_set_log_func(impl_acp_log_func, this);
+	acp_set_log_func(impl_acp_log_func, this->log);
 	acp_set_log_level(6);
 
 	this->device.iface = SPA_INTERFACE_INIT(
