@@ -83,6 +83,7 @@ struct impl {
 	unsigned int have_format:1;
 	unsigned int started:1;
 	unsigned int driver:1;
+	unsigned int async:1;
 };
 
 /** \endcond */
@@ -245,7 +246,7 @@ static int negotiate_buffers(struct impl *this)
 	struct spa_pod *param;
 	int res;
 	bool follower_alloc, conv_alloc;
-	uint32_t i, size, buffers, blocks, align, flags;
+	uint32_t i, size, buffers, blocks, align, flags, stride = 0;
 	uint32_t *aligns;
 	struct spa_data *datas;
 	uint32_t follower_flags, conv_flags;
@@ -303,11 +304,12 @@ static int negotiate_buffers(struct impl *this)
 			SPA_PARAM_BUFFERS_buffers, SPA_POD_Int(&buffers),
 			SPA_PARAM_BUFFERS_blocks,  SPA_POD_Int(&blocks),
 			SPA_PARAM_BUFFERS_size,    SPA_POD_Int(&size),
+			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(&stride),
 			SPA_PARAM_BUFFERS_align,   SPA_POD_Int(&align))) < 0)
 		return res;
 
-	spa_log_debug(this->log, "%p: buffers %d, blocks %d, size %d, align %d %d:%d",
-			this, buffers, blocks, size, align, follower_alloc, conv_alloc);
+	spa_log_debug(this->log, "%p: buffers:%d, blocks:%d, size:%d, stride:%d align:%d %d:%d",
+			this, buffers, blocks, size, stride, align, follower_alloc, conv_alloc);
 
 	align = SPA_MAX(align, this->max_align);
 
@@ -514,7 +516,10 @@ static int impl_node_send_command(void *object, const struct spa_command *comman
 		break;
 	case SPA_NODE_COMMAND_Suspend:
 		configure_format(this, 0, NULL);
-		/* fallthrough */
+		SPA_FALLTHROUGH
+	case SPA_NODE_COMMAND_Flush:
+		this->io_buffers.status = SPA_STATUS_OK;
+		SPA_FALLTHROUGH
 	case SPA_NODE_COMMAND_Pause:
 		this->started = false;
 		break;
@@ -610,6 +615,8 @@ static void follower_info(void *data, const struct spa_node_info *info)
 {
 	struct impl *this = data;
 
+	this->async = (info->flags & SPA_NODE_FLAG_ASYNC) != 0;
+
 	if (info->max_input_ports > 0)
 		this->direction = SPA_DIRECTION_INPUT;
         else
@@ -888,31 +895,85 @@ impl_node_port_reuse_buffer(void *object, uint32_t port_id, uint32_t buffer_id)
 static int impl_node_process(void *object)
 {
 	struct impl *this = object;
-	int status = 0;
+	int status = 0, fstatus;
 
 	spa_log_trace_fp(this->log, "%p: process convert:%p driver:%d",
 			this, this->convert, this->driver);
 
 	if (this->direction == SPA_DIRECTION_INPUT) {
-		if (this->convert)
-			status = spa_node_process(this->convert);
-	}
-
-	if (status >= 0)
-		status = spa_node_process(this->follower);
-
-	if (this->direction == SPA_DIRECTION_OUTPUT &&
-	    !this->driver && this->convert) {
-		while (status > 0) {
-			status = spa_node_process(this->convert);
-			if (status & (SPA_STATUS_HAVE_DATA | SPA_STATUS_DRAINED))
+		/* an input node (sink).
+		 * First we run the converter to process the input for the follower
+		 * then if it produced data, we run the follower. */
+		while (true) {
+			status = this->convert ? spa_node_process(this->convert) : 0;
+			/* schedule the follower when the converter needed
+			 * a recycled buffer */
+			if (status == -EPIPE || status == 0)
+				status = SPA_STATUS_HAVE_DATA;
+			else if (status < 0)
 				break;
-			if (status & SPA_STATUS_NEED_DATA) {
-				status = spa_node_process(this->follower);
-				if (status & SPA_STATUS_NEED_DATA)
+
+			if (status & (SPA_STATUS_HAVE_DATA | SPA_STATUS_DRAINED)) {
+				/* as long as the converter produced something or
+				 * is drained, process the follower. */
+				fstatus = spa_node_process(this->follower);
+				if (fstatus < 0) {
+					status = fstatus;
+					break;
+				}
+				/* if the follower doesn't need more data or is
+				 * drained we can stop */
+				if ((fstatus & SPA_STATUS_NEED_DATA) == 0 ||
+				    (fstatus & SPA_STATUS_DRAINED))
 					break;
 			}
+			/* the converter needs more data */
+			if ((status & SPA_STATUS_NEED_DATA))
+				break;
 		}
+	} else if (!this->driver) {
+		bool done;
+		while (true) {
+			/* output node (source). First run the converter to make
+			 * sure we push out any queued data. Then when it needs
+			 * more data, schedule the follower. */
+			status = this->convert ? spa_node_process(this->convert) : 0;
+			if (status == 0)
+				status = SPA_STATUS_NEED_DATA;
+			else if (status < 0)
+				break;
+
+			done = (status & (SPA_STATUS_HAVE_DATA | SPA_STATUS_DRAINED));
+
+			/* when not async, we can return the data when we are done.
+			 * In async mode we might first need to wake up the follower
+			 * to asynchronously provide more data for the next round. */
+			if (!this->async && done)
+				break;
+
+			if (status & SPA_STATUS_NEED_DATA) {
+				/* the converter needs more data, schedule the
+				 * follower */
+				fstatus = spa_node_process(this->follower);
+				if (fstatus < 0) {
+					status = fstatus;
+					break;
+				}
+				/* if the follower didn't produce more data
+				 * we can stop now */
+				if ((fstatus & SPA_STATUS_HAVE_DATA) == 0)
+					break;
+			}
+			/* converter produced something or is drained and we
+			 * scheduled the follower above, we can stop now*/
+			if (done)
+				break;
+		}
+		if (!(status & SPA_STATUS_HAVE_DATA))
+			spa_node_call_xrun(&this->callbacks, 0, 0, NULL);
+
+	} else {
+		status = spa_node_process(this->follower);
 	}
 	spa_log_trace_fp(this->log, "%p: process status:%d", this, status);
 
