@@ -1518,6 +1518,7 @@ static void proxy_link_destroy(void *data)
 	struct link *l = data;
 
 	spa_list_remove(&l->link);
+	spa_hook_remove(&l->listener);
 
 	if (l->endpoint_link) {
 		check_endpoint_link(l->endpoint_link);
@@ -1792,6 +1793,10 @@ static int state_dir(struct sm_media_session *sess)
 				"%s/.config/pipewire-media-session/", home_dir);
 	}
 
+#ifndef O_PATH
+#define O_PATH 0
+#endif
+
 	if ((res = open(impl->state_dir, O_CLOEXEC | O_DIRECTORY | O_PATH)) < 0) {
 		if (errno == ENOENT) {
 			pw_log_info("creating state directory %s", impl->state_dir);
@@ -2032,12 +2037,19 @@ static void session_shutdown(struct impl *impl)
 
 	sm_media_session_emit_destroy(impl);
 
-	if (impl->registry)
+	if (impl->registry) {
+		spa_hook_remove(&impl->registry_listener);
 		pw_proxy_destroy((struct pw_proxy*)impl->registry);
-	if (impl->policy_core)
+	}
+	if (impl->policy_core) {
+		spa_hook_remove(&impl->policy_listener);
+		spa_hook_remove(&impl->proxy_policy_listener);
 		pw_core_disconnect(impl->policy_core);
-	if (impl->monitor_core)
+	}
+	if (impl->monitor_core) {
+		spa_hook_remove(&impl->monitor_listener);
 		pw_core_disconnect(impl->monitor_core);
+	}
 	if (impl->this.info)
 		pw_core_info_free(impl->this.info);
 }
@@ -2121,8 +2133,8 @@ static void show_help(const char *name, const char *enabled, const char *disable
         fprintf(stdout, "%s [options]\n"
              "  -h, --help                            Show this help\n"
              "      --version                         Show version\n"
-             "  -e, --enabled                         Extra enabled options ('%s')\n"
-             "  -d, --disabled                        Extra disabled options ('%s')\n"
+             "  -e, --enabled                         Extra comma separated enabled options ('%s')\n"
+             "  -d, --disabled                        Extra comma separated disabled options ('%s')\n"
              "  -p, --properties                      Extra properties as 'key=value { key=value }'\n",
 	     name, enabled, disabled);
 
