@@ -45,6 +45,7 @@
 #define MAX_IO	32
 
 /** \cond */
+static bool mlock_warned = false;
 
 struct buffer {
 	uint32_t id;
@@ -123,6 +124,7 @@ do_deactivate_link(struct spa_loop *loop,
 static void clear_link(struct node_data *data, struct link *link)
 {
 	struct pw_context *context = data->context;
+	pw_log_debug("link %p", link);
 	pw_loop_invoke(context->data_loop,
 		do_deactivate_link, SPA_ID_INVALID, NULL, 0, true, link);
 	pw_memmap_free(link->map);
@@ -277,7 +279,7 @@ static int client_node_transport(void *object,
 	return 0;
 }
 
-static int add_node_update(struct node_data *data, uint32_t change_mask)
+static int add_node_update(struct node_data *data, uint32_t change_mask, uint32_t info_mask)
 {
 	struct pw_impl_node *node = data->node;
 	struct spa_node_info ni = SPA_NODE_INFO_INIT();
@@ -312,14 +314,11 @@ static int add_node_update(struct node_data *data, uint32_t change_mask)
 	if (change_mask & PW_CLIENT_NODE_UPDATE_INFO) {
 		ni.max_input_ports = node->info.max_input_ports;
 		ni.max_output_ports = node->info.max_output_ports;
-		ni.change_mask = SPA_NODE_CHANGE_MASK_FLAGS |
-			SPA_NODE_CHANGE_MASK_PROPS |
-			SPA_NODE_CHANGE_MASK_PARAMS;
+		ni.change_mask = info_mask;
 		ni.flags = node->spa_flags;
 		ni.props = node->info.props;
 		ni.params = node->info.params;
 		ni.n_params = node->info.n_params;
-
 	}
 
         res = pw_client_node_update(data->client_node,
@@ -647,12 +646,15 @@ client_node_port_use_buffers(void *object,
 		bid->mem = mm;
 
 		if (data->allow_mlock && mlock(mm->ptr, mm->size) < 0)
-			pw_log(data->warn_mlock ? SPA_LOG_LEVEL_WARN : SPA_LOG_LEVEL_DEBUG,
-					"Failed to mlock memory %p %u: %s",
-					mm->ptr, mm->size,
-					errno == ENOMEM ?
-					"This is not a problem but for best performance, "
-					"consider increasing RLIMIT_MEMLOCK" : strerror(errno));
+			if (errno != ENOMEM || !mlock_warned) {
+				pw_log(data->warn_mlock ? SPA_LOG_LEVEL_WARN : SPA_LOG_LEVEL_DEBUG,
+						"Failed to mlock memory %p %u: %s",
+						mm->ptr, mm->size,
+						errno == ENOMEM ?
+						"This is not a problem but for best performance, "
+						"consider increasing RLIMIT_MEMLOCK" : strerror(errno));
+				mlock_warned |= errno == ENOMEM;
+			}
 
 		size = sizeof(struct spa_buffer);
 		for (j = 0; j < buffers[i].buffer->n_metas; j++)
@@ -942,7 +944,10 @@ static void do_node_init(struct node_data *data)
 
 	pw_log_debug("%p: node %p init", data, data->node);
 	add_node_update(data, PW_CLIENT_NODE_UPDATE_PARAMS |
-				PW_CLIENT_NODE_UPDATE_INFO);
+				PW_CLIENT_NODE_UPDATE_INFO,
+				SPA_NODE_CHANGE_MASK_FLAGS |
+				SPA_NODE_CHANGE_MASK_PROPS |
+				SPA_NODE_CHANGE_MASK_PARAMS);
 
 	spa_list_for_each(port, &data->node->input_ports, link) {
 		add_port_update(data, port,
@@ -1001,21 +1006,23 @@ static void node_free(void *data)
 static void node_info_changed(void *data, const struct pw_node_info *info)
 {
 	struct node_data *d = data;
-	uint32_t change_mask;
+	uint32_t change_mask, info_mask;
 
 	pw_log_debug("info changed %p", d);
 
 	if (d->client_node == NULL)
 		return;
 
-	change_mask = 0;
-	if (info->change_mask & PW_NODE_CHANGE_MASK_PROPS)
-		change_mask |= PW_CLIENT_NODE_UPDATE_INFO;
+	change_mask = PW_CLIENT_NODE_UPDATE_INFO;
+	info_mask = SPA_NODE_CHANGE_MASK_FLAGS;
+	if (info->change_mask & PW_NODE_CHANGE_MASK_PROPS) {
+		info_mask |= SPA_NODE_CHANGE_MASK_PROPS;
+	}
 	if (info->change_mask & PW_NODE_CHANGE_MASK_PARAMS) {
 		change_mask |= PW_CLIENT_NODE_UPDATE_PARAMS;
-		change_mask |= PW_CLIENT_NODE_UPDATE_INFO;
+		info_mask |= SPA_NODE_CHANGE_MASK_PARAMS;
 	}
-	add_node_update(d, change_mask);
+	add_node_update(d, change_mask, info_mask);
 }
 
 static void node_port_info_changed(void *data, struct pw_impl_port *port,
