@@ -36,11 +36,14 @@
 #if HAVE_PWD_H
 #include <pwd.h>
 #endif
+#if defined(__FreeBSD__)
+#include <sys/ucred.h>
+#endif
 
 #include <spa/pod/iter.h>
 #include <spa/utils/result.h>
 
-#ifdef HAVE_SYSTEMD_DAEMON
+#ifdef HAVE_SYSTEMD
 #include <systemd/sd-daemon.h>
 #endif
 
@@ -103,7 +106,7 @@ struct client {
 	int ref;
 
 	unsigned int disconnecting:1;
-	unsigned int flushing:1;
+	unsigned int need_flush:1;
 	unsigned int paused:1;
 };
 
@@ -122,7 +125,7 @@ struct server {
 
 	struct pw_loop *loop;
 	struct spa_source *source;
-	struct spa_hook hook;
+	struct spa_source *resume;
 	unsigned int activated:1;
 };
 
@@ -131,6 +134,7 @@ struct client_data {
 	struct spa_hook client_listener;
 
 	struct spa_list protocol_link;
+	struct server *server;
 
 	struct spa_source *source;
 	struct pw_protocol_native_connection *connection;
@@ -254,6 +258,7 @@ static void
 client_busy_changed(void *data, bool busy)
 {
 	struct client_data *c = data;
+	struct server *s = c->server;
 	struct pw_impl_client *client = c->client;
 	uint32_t mask = c->source->mask;
 
@@ -265,7 +270,17 @@ client_busy_changed(void *data, bool busy)
 	pw_loop_update_io(client->context->main_loop, c->source, mask);
 
 	if (!busy)
-		process_messages(c);
+		pw_loop_signal_event(s->loop, s->resume);
+}
+
+static void handle_client_error(struct pw_impl_client *client, int res)
+{
+	if (res == -EPIPE)
+		pw_log_info(NAME" %p: client %p disconnected", client->protocol, client);
+	else
+		pw_log_error(NAME" %p: client %p error %d (%s)", client->protocol,
+				client, res, spa_strerror(res));
+	pw_impl_client_destroy(client);
 }
 
 static void
@@ -283,28 +298,22 @@ connection_data(void *data, int fd, uint32_t mask)
 		res = -EIO;
 		goto error;
 	}
-	if (mask & SPA_IO_OUT) {
-		res = pw_protocol_native_connection_flush(this->connection);
-		if (res >= 0) {
-			int mask = this->source->mask;
-			SPA_FLAG_CLEAR(mask, SPA_IO_OUT);
-			pw_loop_update_io(client->context->main_loop,
-					this->source, mask);
-		} else if (res != EAGAIN)
-			goto error;
-	}
 	if (mask & SPA_IO_IN) {
 		if ((res = process_messages(this)) < 0)
 			goto error;
 	}
+	if (mask & SPA_IO_OUT || this->need_flush) {
+		this->need_flush = false;
+		res = pw_protocol_native_connection_flush(this->connection);
+		if (res >= 0) {
+			pw_loop_update_io(client->context->main_loop,
+					this->source, this->source->mask & ~SPA_IO_OUT);
+		} else if (res != -EAGAIN)
+			goto error;
+	}
 	return;
 error:
-	if (res == -EPIPE)
-		pw_log_info(NAME" %p: client %p disconnected", client->protocol, client);
-	else
-		pw_log_error(NAME" %p: client %p error %d (%s)", client->protocol,
-				client, res, spa_strerror(res));
-	pw_impl_client_destroy(client);
+	handle_client_error(client, res);
 }
 
 static void client_free(void *data)
@@ -315,6 +324,8 @@ static void client_free(void *data)
 	pw_log_debug(NAME" %p: free", this);
 	spa_list_remove(&this->protocol_link);
 
+	spa_hook_remove(&this->client_listener);
+
 	if (this->source)
 		pw_loop_destroy_source(client->context->main_loop, this->source);
 	if (this->connection)
@@ -329,6 +340,12 @@ static const struct pw_impl_client_events client_events = {
 	.busy_changed = client_busy_changed,
 };
 
+static void on_server_connection_destroy(void *data)
+{
+	struct client_data *this = data;
+	spa_hook_remove(&this->conn_listener);
+}
+
 static void on_start(void *data, uint32_t version)
 {
 	struct client_data *this = data;
@@ -349,9 +366,25 @@ static void on_start(void *data, uint32_t version)
 	return;
 }
 
+static void on_server_need_flush(void *data)
+{
+	struct client_data *this = data;
+	struct pw_impl_client *client = this->client;
+
+	pw_log_debug("need flush");
+	this->need_flush = true;
+
+	if (this->source && !(this->source->mask & SPA_IO_OUT)) {
+		pw_loop_update_io(client->context->main_loop,
+				this->source, this->source->mask | SPA_IO_OUT);
+	}
+}
+
 static const struct pw_protocol_native_connection_events server_conn_events = {
 	PW_VERSION_PROTOCOL_NATIVE_CONNECTION_EVENTS,
+	.destroy = on_server_connection_destroy,
 	.start = on_start,
+	.need_flush = on_server_need_flush,
 };
 
 static struct client_data *client_new(struct server *s, int fd)
@@ -361,6 +394,9 @@ static struct client_data *client_new(struct server *s, int fd)
 	struct pw_protocol *protocol = s->this.protocol;
 	socklen_t len;
 	struct ucred ucred;
+#if defined(__FreeBSD__)
+	struct xucred xucred;
+#endif
 	struct pw_context *context = protocol->context;
 	struct pw_properties *props;
 	char buffer[1024];
@@ -371,7 +407,7 @@ static struct client_data *client_new(struct server *s, int fd)
 	if (props == NULL)
 		goto exit;
 
-#ifndef __FreeBSD__
+#if defined(__linux__)
 	len = sizeof(ucred);
 	if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &ucred, &len) < 0) {
 		pw_log_warn("server %p: no peercred: %m", s);
@@ -383,12 +419,26 @@ static struct client_data *client_new(struct server *s, int fd)
 
 	len = sizeof(buffer);
 	if (getsockopt(fd, SOL_SOCKET, SO_PEERSEC, buffer, &len) < 0) {
-		pw_log_warn("server %p: no peersec: %m", s);
+		if (errno == ENOPROTOOPT)
+			pw_log_info("server %p: security label not available", s);
+		else
+			pw_log_warn("server %p: security label error: %m", s);
 	} else {
 		/* buffer is not null terminated, must use length explicitly */
 		pw_properties_setf(props, PW_KEY_SEC_LABEL, "%.*s",
 				(int)len, buffer);
 	}
+#elif defined(__FreeBSD__)
+	len = sizeof(xucred);
+	if (getsockopt(fd, 0, LOCAL_PEERCRED, &xucred, &len) < 0) {
+		pw_log_warn("server %p: no peercred: %m", s);
+	} else {
+#if __FreeBSD__ >= 13
+		pw_properties_setf(props, PW_KEY_SEC_PID, "%d", xucred.cr_pid);
+#endif
+		pw_properties_setf(props, PW_KEY_SEC_UID, "%d", xucred.cr_uid);
+		pw_properties_setf(props, PW_KEY_SEC_GID, "%d", xucred.cr_gid);
+	}
 #endif
 
 	pw_properties_setf(props, PW_KEY_MODULE_ID, "%d", d->module->global->id);
@@ -402,6 +452,7 @@ static struct client_data *client_new(struct server *s, int fd)
 	this = pw_impl_client_get_user_data(client);
 	spa_list_append(&s->this.client_list, &this->protocol_link);
 
+	this->server = s;
 	this->client = client;
 	this->source = pw_loop_add_io(pw_context_get_main_loop(context),
 				      fd, SPA_IO_ERR | SPA_IO_HUP, true,
@@ -567,7 +618,7 @@ static int add_socket(struct pw_protocol *protocol, struct server *s)
 	int fd = -1, res;
 	bool activated = false;
 
-#ifdef HAVE_SYSTEMD_DAEMON
+#ifdef HAVE_SYSTEMD
 	{
 		int i, n = sd_listen_fds(0);
 		for (i = 0; i < n; ++i) {
@@ -727,59 +778,72 @@ on_remote_data(void *data, int fd, uint32_t mask)
 {
 	struct client *impl = data;
 	struct pw_core *this = impl->this.core;
+	struct pw_proxy *core_proxy = (struct pw_proxy*)this;
 	struct pw_protocol_native_connection *conn = impl->connection;
 	struct pw_context *context = pw_core_get_context(this);
 	struct pw_loop *loop = pw_context_get_main_loop(context);
 	int res;
 
-        if (mask & (SPA_IO_ERR | SPA_IO_HUP)) {
+	core_proxy->refcount++;
+	impl->ref++;
+
+	if (mask & (SPA_IO_ERR | SPA_IO_HUP)) {
 		res = -EPIPE;
 		goto error;
 	}
-	if (mask & SPA_IO_OUT) {
-		res = pw_protocol_native_connection_flush(conn);
-		if (res >= 0) {
-			int mask = impl->source->mask;
-			SPA_FLAG_CLEAR(mask, SPA_IO_OUT);
-			pw_loop_update_io(loop,
-					impl->source, mask);
-			impl->flushing = false;
-		} else if (res != EAGAIN)
-			goto error;
-	}
-
-        if (mask & SPA_IO_IN) {
+	if (mask & SPA_IO_IN) {
 		if ((res = process_remote(impl)) < 0)
 			goto error;
 	}
+	if (mask & SPA_IO_OUT || impl->need_flush) {
+		impl->need_flush = false;
+		res = pw_protocol_native_connection_flush(conn);
+		if (res >= 0) {
+			pw_loop_update_io(loop, impl->source,
+					impl->source->mask & ~SPA_IO_OUT);
+		} else if (res != -EAGAIN)
+			goto error;
+	}
+
+done:
+	client_unref(impl);
+	pw_proxy_unref(core_proxy);
 	return;
 error:
 	pw_log_debug(NAME" %p: got connection error %d (%s)", impl, res, spa_strerror(res));
-	pw_proxy_notify((struct pw_proxy*)this,
-			struct pw_core_events, error, 0, 0,
-			this->recv_seq, res, "connection error");
 	if (impl->source) {
 		pw_loop_destroy_source(loop, impl->source);
 		impl->source = NULL;
 	}
+	pw_proxy_notify(core_proxy,
+			struct pw_core_events, error, 0, 0,
+			this->recv_seq, res, "connection error");
+	goto done;
 }
 
-static void on_need_flush(void *data)
+static void on_client_connection_destroy(void *data)
+{
+	struct client *impl = data;
+	spa_hook_remove(&impl->conn_listener);
+}
+
+static void on_client_need_flush(void *data)
 {
         struct client *impl = data;
 
-	if (!impl->flushing && impl->source) {
-		int mask = impl->source->mask;
-		impl->flushing = true;
-		SPA_FLAG_SET(mask, SPA_IO_OUT);
+	pw_log_debug("need flush");
+	impl->need_flush = true;
+
+	if (impl->source && !(impl->source->mask & SPA_IO_OUT)) {
 		pw_loop_update_io(impl->context->main_loop,
-					impl->source, mask);
+				impl->source, impl->source->mask | SPA_IO_OUT);
 	}
 }
 
 static const struct pw_protocol_native_connection_events client_conn_events = {
 	PW_VERSION_PROTOCOL_NATIVE_CONNECTION_EVENTS,
-	.need_flush = on_need_flush,
+	.destroy = on_client_connection_destroy,
+	.need_flush = on_client_need_flush,
 };
 
 static int impl_connect_fd(struct pw_protocol_client *client, int fd, bool do_close)
@@ -790,7 +854,6 @@ static int impl_connect_fd(struct pw_protocol_client *client, int fd, bool do_cl
 	impl->disconnecting = false;
 
 	pw_protocol_native_connection_set_fd(impl->connection, fd);
-	impl->flushing = true;
 	impl->source = pw_loop_add_io(impl->context->main_loop,
 					fd,
 					SPA_IO_IN | SPA_IO_OUT | SPA_IO_HUP | SPA_IO_ERR,
@@ -971,14 +1034,14 @@ static void destroy_server(struct pw_protocol_server *server)
 	pw_log_debug(NAME" %p: server %p", s->this.protocol, s);
 
 	spa_list_remove(&server->link);
-	spa_hook_remove(&s->hook);
 
 	spa_list_for_each_safe(data, tmp, &server->client_list, protocol_link)
 		pw_impl_client_destroy(data->client);
 
 	if (s->source)
 		pw_loop_destroy_source(s->loop, s->source);
-
+	if (s->resume)
+		pw_loop_destroy_source(s->loop, s->resume);
 	if (s->addr.sun_path[0] && !s->activated)
 		unlink(s->addr.sun_path);
 	if (s->lock_addr[0])
@@ -988,34 +1051,24 @@ static void destroy_server(struct pw_protocol_server *server)
 	free(s);
 }
 
-static void on_before_hook(void *_data)
+static void do_resume(void *_data, uint64_t count)
 {
 	struct server *server = _data;
 	struct pw_protocol_server *this = &server->this;
 	struct client_data *data, *tmp;
 	int res;
 
+	pw_log_debug("flush");
+
 	spa_list_for_each_safe(data, tmp, &this->client_list, protocol_link) {
-		res = pw_protocol_native_connection_flush(data->connection);
-		if (res == -EAGAIN) {
-			int mask = data->source->mask;
-			SPA_FLAG_SET(mask, SPA_IO_OUT);
-			pw_loop_update_io(data->client->context->main_loop,
-					data->source, mask);
-		} else if (res < 0) {
-			pw_log_warn("client %p: could not flush: %s",
-					data->client, spa_strerror(res));
-			pw_impl_client_destroy(data->client);
-		}
-
+		if ((res = process_messages(data)) < 0)
+			goto error;
 	}
+	return;
+error:
+	handle_client_error(data->client, res);
 }
 
-static const struct spa_loop_control_hooks impl_hooks = {
-	SPA_VERSION_LOOP_CONTROL_HOOKS,
-	.before = on_before_hook,
-};
-
 static const char *
 get_server_name(const struct spa_dict *props)
 {
@@ -1036,7 +1089,6 @@ create_server(struct pw_protocol *protocol,
                 const struct spa_dict *props)
 {
 	struct pw_protocol_server *this;
-	struct pw_context *context = protocol->context;
 	struct server *s;
 
 	if ((s = calloc(1, sizeof(struct server))) == NULL)
@@ -1052,8 +1104,6 @@ create_server(struct pw_protocol *protocol,
 
 	spa_list_append(&protocol->server_list, &this->link);
 
-	pw_loop_add_hook(pw_context_get_main_loop(context), &s->hook, &impl_hooks, s);
-
 	pw_log_debug(NAME" %p: created server %p", protocol, this);
 
 	return s;
@@ -1085,6 +1135,9 @@ impl_add_server(struct pw_protocol *protocol,
 	if ((res = add_socket(protocol, s)) < 0)
 		goto error;
 
+	if ((s->resume = pw_loop_add_event(s->loop, do_resume, s)) == NULL)
+		goto error;
+
 	pw_log_info(NAME" %p: Listening on '%s'", protocol, name);
 
 	return this;
