@@ -51,6 +51,7 @@
 #define MAX_PORTS	1024
 
 static float empty[MAX_SAMPLES];
+static bool mlock_warned = false;
 
 struct buffer {
 	struct pw_buffer this;
@@ -139,6 +140,8 @@ struct filter {
 	struct data data;
 	uintptr_t seq;
 	struct pw_time time;
+	uint64_t base_pos;
+	uint32_t clock_id;
 
 	unsigned int disconnecting:1;
 	unsigned int disconnect_core:1;
@@ -553,12 +556,15 @@ static int map_data(struct filter *impl, struct spa_data *data, int prot)
 			range.offset, range.size, data->data);
 
 	if (impl->allow_mlock && mlock(data->data, data->maxsize) < 0) {
-		pw_log(impl->warn_mlock ? SPA_LOG_LEVEL_WARN : SPA_LOG_LEVEL_DEBUG,
-				NAME" %p: Failed to mlock memory %p %u: %s", impl,
-						data->data, data->maxsize,
-						errno == ENOMEM ?
-						"This is not a problem but for best performance, "
-						"consider increasing RLIMIT_MEMLOCK" : strerror(errno));
+		if (errno != ENOMEM || !mlock_warned) {
+			pw_log(impl->warn_mlock ? SPA_LOG_LEVEL_WARN : SPA_LOG_LEVEL_DEBUG,
+					NAME" %p: Failed to mlock memory %p %u: %s", impl,
+					data->data, data->maxsize,
+					errno == ENOMEM ?
+					"This is not a problem but for best performance, "
+					"consider increasing RLIMIT_MEMLOCK" : strerror(errno));
+			mlock_warned |= errno == ENOMEM;
+		}
 	}
 	return 0;
 }
@@ -733,11 +739,15 @@ static int impl_port_reuse_buffer(void *object, uint32_t port_id, uint32_t buffe
 static inline void copy_position(struct filter *impl)
 {
 	struct spa_io_position *p = impl->rt.position;
-	if (p != NULL) {
+	if (SPA_UNLIKELY(p != NULL)) {
 		SEQ_WRITE(impl->seq);
 		impl->time.now = p->clock.nsec;
 		impl->time.rate = p->clock.rate;
-		impl->time.ticks = p->clock.position;
+		if (SPA_UNLIKELY(impl->clock_id != p->clock.id)) {
+			impl->base_pos = p->clock.position - impl->time.ticks;
+			impl->clock_id = p->clock.id;
+		}
+		impl->time.ticks = p->clock.position - impl->base_pos;
 		impl->time.delay = p->clock.delay;
 		SEQ_WRITE(impl->seq);
 	}
@@ -1098,7 +1108,8 @@ void pw_filter_destroy(struct pw_filter *filter)
 
 	pw_filter_emit_destroy(filter);
 
-	pw_filter_disconnect(filter);
+	if (!impl->disconnecting)
+		pw_filter_disconnect(filter);
 
 	spa_list_consume(p, &impl->port_list, link)
 		pw_filter_remove_port(p->user_data);
@@ -1115,6 +1126,9 @@ void pw_filter_destroy(struct pw_filter *filter)
 
 	pw_properties_free(filter->properties);
 
+	spa_hook_list_clean(&impl->hooks);
+	spa_hook_list_clean(&filter->listener_list);
+
 	free(filter->name);
 
 	if (impl->data.context)
