@@ -200,15 +200,14 @@ static int start_node(struct pw_impl_node *this)
 	return res;
 }
 
-static void emit_info_changed(struct pw_impl_node *node)
+static void emit_info_changed(struct pw_impl_node *node, bool flags_changed)
 {
-
-	if (node->info.change_mask == 0)
+	if (node->info.change_mask == 0 && !flags_changed)
 		return;
 
 	pw_impl_node_emit_info_changed(node, &node->info);
 
-	if (node->global) {
+	if (node->global && node->info.change_mask != 0) {
 		struct pw_resource *resource;
 		spa_list_for_each(resource, &node->global->resource_list, link)
 			pw_node_resource_info(resource, &node->info);
@@ -298,12 +297,12 @@ static void node_update_state(struct pw_impl_node *node, enum pw_node_state stat
 	node->info.state = state;
 	impl->pending = state;
 
-	if (old == state)
-		return;
-
 	pw_log_debug(NAME" %p: (%s) %s -> %s (%s)", node, node->name,
 		     pw_node_state_as_string(old), pw_node_state_as_string(state), error);
 
+	if (old == state)
+		return;
+
 	if (state == PW_NODE_STATE_ERROR) {
 		pw_log_error("(%s-%u) %s -> error (%s)", node->name, node->info.id,
 		     pw_node_state_as_string(old), error);
@@ -323,7 +322,7 @@ static void node_update_state(struct pw_impl_node *node, enum pw_node_state stat
 	pw_impl_node_emit_state_changed(node, old, state, error);
 
 	node->info.change_mask |= PW_NODE_CHANGE_MASK_STATE;
-	emit_info_changed(node);
+	emit_info_changed(node, false);
 
 	if (state == PW_NODE_STATE_ERROR && node->global) {
 		struct pw_resource *resource;
@@ -510,6 +509,8 @@ static void resource_destroy(void *data)
 {
 	struct resource_data *d = data;
 	remove_busy_resource(d);
+	spa_hook_remove(&d->resource_listener);
+	spa_hook_remove(&d->object_listener);
 }
 
 static void resource_pong(void *data, int seq)
@@ -1221,7 +1222,7 @@ SPA_EXPORT
 int pw_impl_node_update_properties(struct pw_impl_node *node, const struct spa_dict *dict)
 {
 	int changed = update_properties(node, dict);
-	emit_info_changed(node);
+	emit_info_changed(node, false);
 	return changed;
 }
 
@@ -1229,7 +1230,7 @@ static void node_info(void *data, const struct spa_node_info *info)
 {
 	struct pw_impl_node *node = data;
 	uint32_t changed_ids[MAX_PARAMS], n_changed_ids = 0;
-	bool recalc = false;
+	bool flags_changed = false;
 
 	node->info.max_input_ports = info->max_input_ports;
 	node->info.max_output_ports = info->max_output_ports;
@@ -1239,9 +1240,10 @@ static void node_info(void *data, const struct spa_node_info *info)
 			info->max_output_ports);
 
 	if (info->change_mask & SPA_NODE_CHANGE_MASK_FLAGS) {
-		if ((recalc = node->spa_flags != info->flags)) {
+		if (node->spa_flags != info->flags) {
+			flags_changed = node->spa_flags != 0;
+			pw_log_debug(NAME" %p: flags %"PRIu64"->%"PRIu64, node, node->spa_flags, info->flags);
 			node->spa_flags = info->flags;
-			node->info.change_mask |= PW_NODE_CHANGE_MASK_PROPS;
 		}
 	}
 	if (info->change_mask & SPA_NODE_CHANGE_MASK_PROPS) {
@@ -1266,13 +1268,13 @@ static void node_info(void *data, const struct spa_node_info *info)
 			node->info.params[i] = info->params[i];
 		}
 	}
-	emit_info_changed(node);
+	emit_info_changed(node, flags_changed);
 
 	if (n_changed_ids > 0)
 		emit_params(node, changed_ids, n_changed_ids);
 
-	if (recalc)
-		pw_context_recalc_graph(node->context, "node info changed");
+	if (flags_changed)
+		pw_context_recalc_graph(node->context, "node flags changed");
 }
 
 static void node_port_info(void *data, enum spa_direction direction, uint32_t port_id,
@@ -1678,6 +1680,8 @@ void pw_impl_node_destroy(struct pw_impl_node *node)
 	pw_log_debug(NAME" %p: free", node);
 	pw_impl_node_emit_free(node);
 
+	spa_hook_list_clean(&node->listener_list);
+
 	pw_memblock_unref(node->activation);
 
 	pw_work_queue_destroy(impl->work);
@@ -1921,7 +1925,8 @@ int pw_impl_node_set_state(struct pw_impl_node *node, enum pw_node_state state)
 	struct impl *impl = SPA_CONTAINER_OF(node, struct impl, this);
 	enum pw_node_state old = impl->pending;
 
-	pw_log_debug(NAME" %p: set state %s -> %s, active %d", node,
+	pw_log_debug(NAME" %p: set state (%s) %s -> %s, active %d", node,
+			pw_node_state_as_string(node->info.state),
 			pw_node_state_as_string(old),
 			pw_node_state_as_string(state),
 			node->active);
@@ -1960,9 +1965,11 @@ int pw_impl_node_set_state(struct pw_impl_node *node, enum pw_node_state state)
 	}
 
 	if (old != state) {
-		impl->pending = state;
-		if (impl->pending_id != SPA_ID_INVALID)
+		if (impl->pending_id != SPA_ID_INVALID) {
 			pw_work_queue_cancel(impl->work, node, impl->pending_id);
+			node->info.state = impl->pending;
+		}
+		impl->pending = state;
 		impl->pending_id = pw_work_queue_add(impl->work,
 				node, res, on_state_complete, SPA_INT_TO_PTR(state));
 	}
