@@ -51,6 +51,7 @@ struct resource_data {
 	struct pw_impl_port *port;
 	struct pw_resource *resource;
 
+	struct spa_hook resource_listener;
 	struct spa_hook object_listener;
 
 	uint32_t subscribe_ids[MAX_PARAMS];
@@ -94,7 +95,7 @@ static const char *port_state_as_string(enum pw_impl_port_state state)
 	return "invalid-state";
 }
 
-void pw_impl_port_update_state(struct pw_impl_port *port, enum pw_impl_port_state state, char *error)
+void pw_impl_port_update_state(struct pw_impl_port *port, enum pw_impl_port_state state, int res, char *error)
 {
 	enum pw_impl_port_state old = port->state;
 
@@ -111,6 +112,12 @@ void pw_impl_port_update_state(struct pw_impl_port *port, enum pw_impl_port_stat
 		port_state_as_string(old), port_state_as_string(state), error);
 
 	pw_impl_port_emit_state_changed(port, old, state, error);
+
+	if (state == PW_IMPL_PORT_STATE_ERROR && port->global) {
+		struct pw_resource *resource;
+		spa_list_for_each(resource, &port->global->resource_list, link)
+			pw_resource_error(resource, res, error);
+	}
 }
 
 static int tee_process(void *object)
@@ -725,6 +732,18 @@ static const struct pw_port_methods port_methods = {
 	.enum_params = port_enum_params
 };
 
+static void resource_destroy(void *data)
+{
+	struct resource_data *d = data;
+	spa_hook_remove(&d->resource_listener);
+	spa_hook_remove(&d->object_listener);
+}
+
+static const struct pw_resource_events resource_events = {
+	PW_VERSION_RESOURCE_EVENTS,
+	.destroy = resource_destroy,
+};
+
 static int
 global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
 	       uint32_t version, uint32_t id)
@@ -745,6 +764,9 @@ global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
 	data->port = this;
 	data->resource = resource;
 
+	pw_resource_add_listener(resource,
+			&data->resource_listener,
+			&resource_events, data);
 	pw_resource_add_object_listener(resource,
 			&data->object_listener,
 			&port_methods, data);
@@ -908,7 +930,7 @@ int pw_impl_port_add(struct pw_impl_port *port, struct pw_impl_node *node)
 	pw_loop_invoke(node->data_loop, do_add_port, SPA_ID_INVALID, NULL, 0, false, port);
 
 	if (port->state <= PW_IMPL_PORT_STATE_INIT)
-		pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_CONFIGURE, NULL);
+		pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_CONFIGURE, 0, NULL);
 
 	pw_impl_node_emit_port_added(node, port);
 	emit_info_changed(port);
@@ -996,6 +1018,8 @@ void pw_impl_port_destroy(struct pw_impl_port *port)
 	pw_log_debug(NAME" %p: free", port);
 	pw_impl_port_emit_free(port);
 
+	spa_hook_list_clean(&port->listener_list);
+
 	pw_buffers_clear(&port->buffers);
 	pw_buffers_clear(&port->mix_buffers);
 	free((void*)port->error);
@@ -1183,10 +1207,10 @@ int pw_impl_port_set_param(struct pw_impl_port *port, uint32_t id, uint32_t flag
 		pw_buffers_clear(&port->mix_buffers);
 
 		if (param == NULL || res < 0) {
-			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_CONFIGURE, NULL);
+			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_CONFIGURE, 0, NULL);
 		}
 		else if (!SPA_RESULT_IS_ASYNC(res)) {
-			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_READY, NULL);
+			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_READY, 0, NULL);
 		}
 	}
 	return res;
@@ -1261,7 +1285,7 @@ int pw_impl_port_use_buffers(struct pw_impl_port *port, struct pw_impl_port_mix
 
 	if (n_buffers == 0) {
 		if (port->n_mix == 1)
-			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_READY, NULL);
+			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_READY, 0, NULL);
 	}
 
 	/* first negotiate with the node, this makes it possible to let the
@@ -1272,10 +1296,10 @@ int pw_impl_port_use_buffers(struct pw_impl_port *port, struct pw_impl_port_mix
 		if (res < 0) {
 			pw_log_error(NAME" %p: negotiate buffers on node: %d (%s)",
 				port, res, spa_strerror(res));
-			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_ERROR,
+			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_ERROR, res,
 					strdup("can't negotiate buffers on port"));
 		} else if (n_buffers > 0 && !SPA_RESULT_IS_ASYNC(res)) {
-			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_PAUSED, NULL);
+			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_PAUSED, 0, NULL);
 		}
 	}
 
