@@ -43,11 +43,7 @@
 
 #define NAME "mempool"
 
-#ifndef __FreeBSD__
-#define USE_MEMFD
-#endif
-
-#if defined(USE_MEMFD) && !defined(HAVE_MEMFD_CREATE)
+#if !defined(__FreeBSD__) && !defined(HAVE_MEMFD_CREATE)
 /*
  * No glibc wrappers exist for memfd_create(2), so provide our own.
  *
@@ -60,6 +56,8 @@ static inline int memfd_create(const char *name, unsigned int flags)
 {
 	return syscall(SYS_memfd_create, name, flags);
 }
+
+#define HAVE_MEMFD_CREATE 1
 #endif
 
 /* memfd_create(2) flags */
@@ -98,22 +96,23 @@ static struct spa_list _mempools = SPA_LIST_INIT(&_mempools);
 struct mempool {
 	struct pw_mempool this;
 
-	struct spa_list link;
+	struct spa_list link;		/* link in global _mempools */
 
 	struct spa_hook_list listener_list;
 
-	struct pw_map map;
-	struct spa_list blocks;
+	struct pw_map map;		/* map memblock to id */
+	struct spa_list blocks;		/* list of memblock */
 	uint32_t pagesize;
 };
 
 struct memblock {
 	struct pw_memblock this;
-	struct spa_list link;
-	struct spa_list mappings;
-	struct spa_list maps;
+	struct spa_list link;		/* link in mempool */
+	struct spa_list mappings;	/* list of struct mapping */
+	struct spa_list memmaps;	/* list of struct memmap */
 };
 
+/* a mapped region of a block */
 struct mapping {
 	struct memblock *block;
 	int ref;
@@ -124,6 +123,7 @@ struct mapping {
 	void *ptr;
 };
 
+/* a reference to a (part of a) mapped region */
 struct memmap {
 	struct pw_memmap this;
 	struct mapping *mapping;
@@ -160,8 +160,11 @@ void pw_mempool_clear(struct pw_mempool *pool)
 	struct mempool *impl = SPA_CONTAINER_OF(pool, struct mempool, this);
 	struct memblock *b;
 
+	pw_log_debug(NAME" %p: clear", pool);
+
 	spa_list_consume(b, &impl->blocks, link)
 		pw_memblock_free(&b->this);
+	pw_map_reset(&impl->map);
 }
 
 void pw_mempool_destroy(struct pw_mempool *pool)
@@ -176,6 +179,8 @@ void pw_mempool_destroy(struct pw_mempool *pool)
 
 	spa_list_remove(&impl->link);
 
+	spa_hook_list_clean(&impl->listener_list);
+
 	pw_map_clear(&impl->map);
 	if (pool->props)
 		pw_properties_free(pool->props);
@@ -260,6 +265,9 @@ static struct mapping * memblock_find_mapping(struct memblock *b,
 	struct pw_mempool *pool = b->this.pool;
 
 	spa_list_for_each(m, &b->mappings, link) {
+		pw_log_debug(NAME" %p: check %p offset:(%d <= %d) end:(%d >= %d)",
+				pool, m, m->offset, offset, m->offset + m->size,
+				offset + size);
 		if (m->offset <= offset && (m->offset + m->size) >= (offset + size)) {
 			pw_log_debug(NAME" %p: found %p id:%d fd:%d offs:%d size:%d ref:%d",
 					pool, &b->this, b->this.id, b->this.fd,
@@ -315,25 +323,33 @@ static struct mapping * memblock_map(struct memblock *b,
 	b->this.ref++;
 	spa_list_append(&b->mappings, &m->link);
 
-        pw_log_debug(NAME" %p: fd:%d map:%p ptr:%p (%d %d)", p,
-			b->this.fd, m, m->ptr, offset, size);
+        pw_log_debug(NAME" %p: block:%p fd:%d map:%p ptr:%p (%d %d) block-ref:%d", p, &b->this,
+			b->this.fd, m, m->ptr, offset, size, b->this.ref);
 
 	return m;
 }
 
+static void mapping_free(struct mapping *m)
+{
+	struct memblock *b = m->block;
+	struct mempool *p = SPA_CONTAINER_OF(b->this.pool, struct mempool, this);
+
+        pw_log_debug(NAME" %p: mapping:%p block:%p fd:%d ptr:%p size:%d block-ref:%d",
+			p, m, b, b->this.fd, m->ptr, m->size, b->this.ref);
+
+	if (m->do_unmap)
+		munmap(m->ptr, m->size);
+	spa_list_remove(&m->link);
+	free(m);
+}
+
 static void mapping_unmap(struct mapping *m)
 {
 	struct memblock *b = m->block;
 	struct mempool *p = SPA_CONTAINER_OF(b->this.pool, struct mempool, this);
-
-        pw_log_debug(NAME" %p: mapping:%p fd:%d ptr:%p size:%d block-ref:%d",
-			p, m, b->this.fd, m->ptr, m->size, b->this.ref);
-
-	if (m->do_unmap)
-		munmap(m->ptr, m->size);
-	spa_list_remove(&m->link);
-	free(m);
-
+        pw_log_debug(NAME" %p: mapping:%p block:%p fd:%d ptr:%p size:%d block-ref:%d",
+			p, m, b, b->this.fd, m->ptr, m->size, b->this.ref);
+	mapping_free(m);
 	pw_memblock_unref(&b->this);
 }
 
@@ -349,7 +365,7 @@ struct pw_memmap * pw_memblock_map(struct pw_memblock *block,
 
 	pw_map_range_init(&range, offset, size, p->pagesize);
 
-	m = memblock_find_mapping(b, flags, range.offset, range.size);
+	m = memblock_find_mapping(b, flags, offset, size);
 	if (m == NULL)
 		m = memblock_map(b, flags, range.offset, range.size);
 	if (m == NULL)
@@ -369,13 +385,17 @@ struct pw_memmap * pw_memblock_map(struct pw_memblock *block,
 	mm->this.offset = offset;
 	mm->this.size = size;
 	mm->this.ptr = SPA_MEMBER(m->ptr, range.start, void);
-	if (tag)
+
+        pw_log_debug(NAME" %p: map:%p block:%p fd:%d ptr:%p (%d %d) mapping:%p ref:%d", p,
+			&mm->this, b, b->this.fd, mm->this.ptr, offset, size, m, m->ref);
+
+	if (tag) {
 		memcpy(mm->this.tag, tag, sizeof(mm->this.tag));
+		pw_log_debug(NAME" %p: tag:%d:%d:%d:%d:%d", p,
+			tag[0], tag[1], tag[2], tag[3], tag[4]);
+	}
 
-	spa_list_append(&b->maps, &mm->link);
-
-        pw_log_debug(NAME" %p: map:%p fd:%d ptr:%p (%d %d) mapping:%p ref:%d", p,
-			&mm->this, b->this.fd, mm->this.ptr, offset, size, m, m->ref);
+	spa_list_append(&b->memmaps, &mm->link);
 
 	return &mm->this;
 }
@@ -403,8 +423,8 @@ int pw_memmap_free(struct pw_memmap *map)
 	struct memblock *b = m->block;
 	struct mempool *p = SPA_CONTAINER_OF(b->this.pool, struct mempool, this);
 
-        pw_log_debug(NAME" %p: map:%p fd:%d ptr:%p mapping:%p ref:%d", p,
-			&mm->this, b->this.fd, mm->this.ptr, m, m->ref);
+        pw_log_debug(NAME" %p: map:%p block:%p fd:%d ptr:%p mapping:%p ref:%d", p,
+			&mm->this, b, b->this.fd, mm->this.ptr, m, m->ref);
 
 	spa_list_remove(&mm->link);
 
@@ -454,15 +474,22 @@ struct pw_memblock * pw_mempool_alloc(struct pw_mempool *pool, enum pw_memblock_
 	b->this.type = type;
 	b->this.size = size;
 	spa_list_init(&b->mappings);
-	spa_list_init(&b->maps);
+	spa_list_init(&b->memmaps);
 
-#ifdef USE_MEMFD
+#ifdef HAVE_MEMFD_CREATE
 	b->this.fd = memfd_create("pipewire-memfd", MFD_CLOEXEC | MFD_ALLOW_SEALING);
 	if (b->this.fd == -1) {
 		res = -errno;
 		pw_log_error(NAME" %p: Failed to create memfd: %m", pool);
 		goto error_free;
 	}
+#elif defined(__FreeBSD__)
+	b->this.fd = shm_open(SHM_ANON, O_CREAT | O_RDWR | O_CLOEXEC, 0);
+	if (b->this.fd == -1) {
+		res = -errno;
+		pw_log_error(NAME" %p: Failed to create SHM_ANON fd: %m", pool);
+		goto error_free;
+	}
 #else
 	char filename[] = "/dev/shm/pipewire-tmpfile.XXXXXX";
 	b->this.fd = mkostemp(filename, O_CLOEXEC);
@@ -479,7 +506,7 @@ struct pw_memblock * pw_mempool_alloc(struct pw_mempool *pool, enum pw_memblock_
 		pw_log_warn(NAME" %p: Failed to truncate temporary file: %m", pool);
 		goto error_close;
 	}
-#ifdef USE_MEMFD
+#ifdef HAVE_MEMFD_CREATE
 	if (flags & PW_MEMBLOCK_FLAG_SEAL) {
 		unsigned int seals = F_SEAL_GROW | F_SEAL_SHRINK | F_SEAL_SEAL;
 		if (fcntl(b->this.fd, F_ADD_SEALS, seals) == -1) {
@@ -500,9 +527,10 @@ struct pw_memblock * pw_mempool_alloc(struct pw_mempool *pool, enum pw_memblock_
 
 	b->this.id = pw_map_insert_new(&impl->map, b);
 	spa_list_append(&impl->blocks, &b->link);
-	pw_log_debug(NAME" %p: mem %p alloc id:%d type:%u", pool, &b->this, b->this.id, type);
+	pw_log_debug(NAME" %p: block:%p id:%d type:%u size:%zd", pool, &b->this, b->this.id, type, size);
 
-	pw_mempool_emit_added(impl, &b->this);
+	if (!SPA_FLAG_IS_SET(flags, PW_MEMBLOCK_FLAG_DONT_NOTIFY))
+		pw_mempool_emit_added(impl, &b->this);
 
 	return &b->this;
 
@@ -546,7 +574,7 @@ struct pw_memblock * pw_mempool_import(struct pw_mempool *pool,
 	if (b == NULL)
 		return NULL;
 
-	spa_list_init(&b->maps);
+	spa_list_init(&b->memmaps);
 	spa_list_init(&b->mappings);
 
 	b->this.ref = 1;
@@ -557,10 +585,11 @@ struct pw_memblock * pw_mempool_import(struct pw_mempool *pool,
 	b->this.id = pw_map_insert_new(&impl->map, b);
 	spa_list_append(&impl->blocks, &b->link);
 
-	pw_log_debug(NAME" %p: import %p id:%u flags:%08x type:%u fd:%d",
+	pw_log_debug(NAME" %p: block:%p id:%u flags:%08x type:%u fd:%d",
 			pool, b, b->this.id, flags, type, fd);
 
-	pw_mempool_emit_added(impl, &b->this);
+	if (!SPA_FLAG_IS_SET(flags, PW_MEMBLOCK_FLAG_DONT_NOTIFY))
+		pw_mempool_emit_added(impl, &b->this);
 
 	return &b->this;
 }
@@ -569,6 +598,8 @@ SPA_EXPORT
 struct pw_memblock * pw_mempool_import_block(struct pw_mempool *pool,
 		struct pw_memblock *mem)
 {
+	pw_log_debug(NAME" %p: import block:%p type:%d fd:%d", pool,
+			mem, mem->type, mem->fd);
 	return pw_mempool_import(pool,
 			mem->flags | PW_MEMBLOCK_FLAG_DONT_CLOSE,
 			mem->type, mem->fd);
@@ -581,7 +612,6 @@ struct pw_memmap * pw_mempool_import_map(struct pw_mempool *pool,
 	struct pw_memblock *old, *block;
 	struct memblock *b;
 	struct pw_memmap *map;
-	struct mapping *m;
 	uint32_t offset;
 
 	old = pw_mempool_find_ptr(other, data);
@@ -595,6 +625,8 @@ struct pw_memmap * pw_mempool_import_map(struct pw_mempool *pool,
 		return NULL;
 
 	if (block->ref == 1) {
+		struct mapping *m;
+
 		b = SPA_CONTAINER_OF(block, struct memblock, this);
 
 		m = calloc(1, sizeof(struct mapping));
@@ -607,6 +639,8 @@ struct pw_memmap * pw_mempool_import_map(struct pw_mempool *pool,
 		m->offset = old->map->offset;
 		m->size = old->map->size;
 		spa_list_append(&b->mappings, &m->link);
+		pw_log_debug(NAME" %p: mapping:%p block:%p offset:%u size:%u ref:%u",
+				pool, m, block, m->offset, m->size, block->ref);
 	} else {
 		block->ref--;
 	}
@@ -618,11 +652,13 @@ struct pw_memmap * pw_mempool_import_map(struct pw_mempool *pool,
 	if (map == NULL)
 		return NULL;
 
+	pw_log_debug(NAME" %p: from pool:%p block:%p id:%u data:%p size:%u ref:%d",
+			pool, other, block, block->id, data, size, block->ref);
+
 	return map;
 }
 
-
-int pw_mempool_unref_id(struct pw_mempool *pool, uint32_t id)
+int pw_mempool_remove_id(struct pw_mempool *pool, uint32_t id)
 {
 	struct mempool *impl = SPA_CONTAINER_OF(pool, struct mempool, this);
 	struct memblock *b;
@@ -631,7 +667,13 @@ int pw_mempool_unref_id(struct pw_mempool *pool, uint32_t id)
 	if (b == NULL)
 		return -ENOENT;
 
+	pw_log_debug(NAME" %p: block:%p id:%d fd:%d ref:%d",
+			pool, b, id, b->this.fd, b->this.ref);
+
+	b->this.id = SPA_ID_INVALID;
+	pw_map_remove(&impl->map, id);
 	pw_memblock_unref(&b->this);
+
 	return 0;
 }
 
@@ -646,24 +688,32 @@ void pw_memblock_free(struct pw_memblock *block)
 	struct pw_mempool *pool = block->pool;
 	struct mempool *impl = SPA_CONTAINER_OF(pool, struct mempool, this);
 	struct memmap *mm;
+	struct mapping *m;
 
 	spa_return_if_fail(block != NULL);
 
-	pw_log_debug(NAME" %p: free mem %p id:%d fd:%d ref:%d",
+	pw_log_debug(NAME" %p: block:%p id:%d fd:%d ref:%d",
 			pool, block, block->id, block->fd, block->ref);
 
 	block->ref++;
 	if (block->map)
 		block->ref++;
 
-	pw_map_remove(&impl->map, block->id);
+	if (block->id != SPA_ID_INVALID)
+		pw_map_remove(&impl->map, block->id);
 	spa_list_remove(&b->link);
 
-	pw_mempool_emit_removed(impl, block);
+	if (!SPA_FLAG_IS_SET(block->flags, PW_MEMBLOCK_FLAG_DONT_NOTIFY))
+		pw_mempool_emit_removed(impl, block);
 
-	spa_list_consume(mm, &b->maps, link)
+	spa_list_consume(mm, &b->memmaps, link)
 		pw_memmap_free(&mm->this);
 
+	spa_list_consume(m, &b->mappings, link) {
+		pw_log_warn(NAME" %p: stray mapping:%p", pool, m);
+		mapping_free(m);
+	}
+
 	if (block->fd != -1 && !(block->flags & PW_MEMBLOCK_FLAG_DONT_CLOSE)) {
 		pw_log_debug(NAME" %p: close fd:%d", pool, block->fd);
 		close(block->fd);
@@ -681,8 +731,8 @@ struct pw_memblock * pw_mempool_find_ptr(struct pw_mempool *pool, const void *pt
 	spa_list_for_each(b, &impl->blocks, link) {
 		spa_list_for_each(m, &b->mappings, link) {
 			if (ptr >= m->ptr && ptr < SPA_MEMBER(m->ptr, m->size, void)) {
-				pw_log_debug(NAME" %p: found %p id:%d for %p", pool,
-						m->block, b->this.id, ptr);
+				pw_log_debug(NAME" %p: block:%p id:%d for %p", pool,
+						b, b->this.id, ptr);
 				return &b->this;
 			}
 		}
@@ -697,7 +747,7 @@ struct pw_memblock * pw_mempool_find_id(struct pw_mempool *pool, uint32_t id)
 	struct memblock *b;
 
 	b = pw_map_lookup(&impl->map, id);
-	pw_log_debug(NAME" %p: found %p for %d", pool, b, id);
+	pw_log_debug(NAME" %p: block:%p for %d", pool, b, id);
 	if (b == NULL)
 		return NULL;
 
@@ -723,10 +773,11 @@ struct pw_memmap * pw_mempool_find_tag(struct pw_mempool *pool, uint32_t tag[5],
 	struct memblock *b;
 	struct memmap *mm;
 
-	pw_log_debug(NAME" %p: find tag %zd", pool, size);
+	pw_log_debug(NAME" %p: find tag %d:%d:%d:%d:%d size:%zd", pool,
+			tag[0], tag[1], tag[2], tag[3], tag[4], size);
 
 	spa_list_for_each(b, &impl->blocks, link) {
-		spa_list_for_each(mm, &b->maps, link) {
+		spa_list_for_each(mm, &b->memmaps, link) {
 			if (memcmp(tag, mm->this.tag, size) == 0) {
 				pw_log_debug(NAME" %p: found %p", pool, mm);
 				return &mm->this;
