@@ -49,6 +49,8 @@
 #define MASK_BUFFERS	(MAX_BUFFERS-1)
 #define MAX_PORTS	1
 
+static bool mlock_warned = false;
+
 struct buffer {
 	struct pw_buffer this;
 	uint32_t id;
@@ -81,6 +83,7 @@ struct param {
 struct control {
 	uint32_t id;
 	uint32_t type;
+	uint32_t container;
 	struct spa_list link;
 	struct pw_stream_control control;
 	struct spa_pod *info;
@@ -138,7 +141,7 @@ struct stream {
 	unsigned int draining:1;
 	unsigned int drained:1;
 	unsigned int allow_mlock:1;
-	unsigned int warn_mlock:1;
+	unsigned int process_rt:1;
 };
 
 static int get_param_index(uint32_t id)
@@ -321,14 +324,12 @@ do_call_process(struct spa_loop *loop,
 static void call_process(struct stream *impl)
 {
 	struct pw_stream *stream = &impl->this;
-	pw_log_trace(NAME" %p: call process", impl);
-	if (SPA_FLAG_IS_SET(impl->flags, PW_STREAM_FLAG_RT_PROCESS)) {
+	pw_log_trace(NAME" %p: call process rt:%u", impl, impl->process_rt);
+	if (impl->process_rt)
 		pw_stream_emit_process(stream);
-	}
-	else {
+	else
 		pw_loop_invoke(impl->context->main_loop,
 			do_call_process, 1, NULL, 0, false, impl);
-	}
 }
 
 static int
@@ -387,6 +388,7 @@ static int impl_send_command(void *object, const struct spa_command *command)
 
 	switch (SPA_NODE_COMMAND_ID(command)) {
 	case SPA_NODE_COMMAND_Suspend:
+	case SPA_NODE_COMMAND_Flush:
 	case SPA_NODE_COMMAND_Pause:
 		pw_loop_invoke(impl->context->main_loop,
 			NULL, 0, NULL, 0, false, impl);
@@ -429,7 +431,13 @@ static void emit_node_info(struct stream *d)
 		info.max_output_ports = 1;
 	}
 	info.change_mask |= SPA_NODE_CHANGE_MASK_FLAGS;
+	/* we're always RT safe, if the stream was marked RT_PROCESS,
+	 * the callback must be RT safe */
 	info.flags = SPA_NODE_FLAG_RT;
+	/* if the callback was not marked RT_PROCESS, we will offload
+	 * the process callback in the main thread and we are ASYNC */
+	if (!d->process_rt)
+		info.flags |= SPA_NODE_FLAG_ASYNC;
 	spa_node_emit_info(&d->hooks, &info);
 }
 
@@ -556,12 +564,15 @@ static int map_data(struct stream *impl, struct spa_data *data, int prot)
 			range.offset, range.size, data->data);
 
 	if (impl->allow_mlock && mlock(data->data, data->maxsize) < 0) {
-		pw_log(impl->warn_mlock ? SPA_LOG_LEVEL_WARN : SPA_LOG_LEVEL_DEBUG,
-				NAME" %p: Failed to mlock memory %p %u: %s", impl,
-				data->data, data->maxsize,
-				errno == ENOMEM ?
-				"This is not a problem but for best performance, "
-				"consider increasing RLIMIT_MEMLOCK" : strerror(errno));
+		if (errno != ENOMEM || !mlock_warned) {
+			pw_log(impl->process_rt ? SPA_LOG_LEVEL_WARN : SPA_LOG_LEVEL_DEBUG,
+					NAME" %p: Failed to mlock memory %p %u: %s", impl,
+					data->data, data->maxsize,
+					errno == ENOMEM ?
+					"This is not a problem but for best performance, "
+					"consider increasing RLIMIT_MEMLOCK" : strerror(errno));
+			mlock_warned |= errno == ENOMEM;
+		}
 	}
 	return 0;
 }
@@ -747,28 +758,24 @@ static int impl_node_process_input(void *object)
 	pw_log_trace(NAME" %p: process in status:%d id:%d ticks:%"PRIu64" delay:%"PRIi64,
 			stream, io->status, io->buffer_id, impl->time.ticks, impl->time.delay);
 
-	if (io->status != SPA_STATUS_HAVE_DATA)
-		goto done;
-
-	if ((b = get_buffer(stream, io->buffer_id)) == NULL)
-		goto done;
-
-	/* push new buffer */
-	if (push_queue(impl, &impl->dequeued, b) == 0)
-		call_process(impl);
-
-done:
-	copy_position(impl, impl->dequeued.incount);
-
-	/* pop buffer to recycle */
-	if ((b = pop_queue(impl, &impl->queued))) {
-		pw_log_trace(NAME" %p: recycle buffer %d", stream, b->id);
+	if (io->status == SPA_STATUS_HAVE_DATA &&
+	    (b = get_buffer(stream, io->buffer_id)) != NULL) {
+		/* push new buffer */
+		if (push_queue(impl, &impl->dequeued, b) == 0) {
+			copy_position(impl, impl->dequeued.incount);
+			call_process(impl);
+		}
 	}
-
-	io->buffer_id = b ? b->id : SPA_ID_INVALID;
-	io->status = SPA_STATUS_NEED_DATA;
-
-	return SPA_STATUS_HAVE_DATA;
+	if (io->status != SPA_STATUS_NEED_DATA) {
+		/* pop buffer to recycle */
+		if ((b = pop_queue(impl, &impl->queued))) {
+			pw_log_trace(NAME" %p: recycle buffer %d", stream, b->id);
+		} else if (io->status == -EPIPE)
+			return io->status;
+		io->buffer_id = b ? b->id : SPA_ID_INVALID;
+		io->status = SPA_STATUS_NEED_DATA;
+	}
+	return SPA_STATUS_NEED_DATA | SPA_STATUS_HAVE_DATA;
 }
 
 static int impl_node_process_output(void *object)
@@ -807,16 +814,26 @@ again:
 		}
 	}
 
-	if (!impl->draining &&
-	    !SPA_FLAG_IS_SET(impl->flags, PW_STREAM_FLAG_DRIVER) &&
-	    spa_ringbuffer_get_read_index(&impl->dequeued.ring, &index) > 0) {
-		call_process(impl);
-		if (spa_ringbuffer_get_read_index(&impl->queued.ring, &index) > 0 &&
-		    io->status == SPA_STATUS_NEED_DATA)
-			goto again;
-	}
 	copy_position(impl, impl->queued.outcount);
 
+	if (!impl->draining &&
+	    !SPA_FLAG_IS_SET(impl->flags, PW_STREAM_FLAG_DRIVER)) {
+		/* we're not draining, not a driver check if we need to get
+		 * more buffers */
+		if (!impl->process_rt) {
+			/* not realtime and we have a free buffer, trigger process so that we have
+			 * data in the next round. */
+			if (spa_ringbuffer_get_read_index(&impl->dequeued.ring, &index) > 0)
+				call_process(impl);
+		} else if (io->status == SPA_STATUS_NEED_DATA) {
+			/* realtime and we don't have a buffer, trigger process and try
+			 * again when there is something in the queue now */
+			call_process(impl);
+			if (spa_ringbuffer_get_read_index(&impl->queued.ring, &index) > 0)
+				goto again;
+		}
+	}
+
 	pw_log_trace(NAME" %p: res %d", stream, res);
 
 	return res;
@@ -895,6 +912,7 @@ static int node_event_param(void *object, int seq,
 		const struct spa_pod *type, *pod;
 		uint32_t iid, choice, n_vals;
 		float *vals, bool_range[3] = { 1.0, 0.0, 1.0 };
+		const struct spa_type_info *tinfo;
 
 		if (spa_pod_parse_object(param,
 					SPA_TYPE_OBJECT_PropInfo, NULL,
@@ -936,6 +954,9 @@ static int node_event_param(void *object, int seq,
 		else
 			return -ENOTSUP;
 
+		tinfo = spa_debug_type_find(spa_type_props, iid);
+		c->container = tinfo ? tinfo->parent : c->type;
+
 		switch (choice) {
 		case SPA_CHOICE_None:
 			if (n_vals < 1)
@@ -959,8 +980,8 @@ static int node_event_param(void *object, int seq,
 		}
 
 		c->id = iid;
-		pw_log_debug(NAME" %p: add control %d (%s) (def:%f min:%f max:%f)",
-				stream, c->id, c->control.name,
+		pw_log_debug(NAME" %p: add control %d (%s) container:%d (def:%f min:%f max:%f)",
+				stream, c->id, c->control.name, c->container,
 				c->control.def, c->control.min, c->control.max);
 		break;
 	}
@@ -982,19 +1003,28 @@ static int node_event_param(void *object, int seq,
 			if (c == NULL)
 				continue;
 
-			if (spa_pod_get_float(&prop->value, &value.f) == 0) {
+			switch (c->container) {
+			case SPA_TYPE_Float:
+				if (spa_pod_get_float(&prop->value, &value.f) < 0)
+					continue;
 				n_values = 1;
 				values = &value.f;
-			} else if (spa_pod_get_bool(&prop->value, &value.b) == 0) {
+				break;
+			case SPA_TYPE_Bool:
+				if (spa_pod_get_bool(&prop->value, &value.b) < 0)
+					continue;
 				value.f = value.b ? 1.0 : 0.0;
 				n_values = 1;
 				values = &value.f;
-			} else if ((values = spa_pod_get_array(&prop->value, &n_values))) {
-				if (!spa_pod_is_float(SPA_POD_ARRAY_CHILD(&prop->value)))
+				break;
+			case SPA_TYPE_Array:
+				if ((values = spa_pod_get_array(&prop->value, &n_values)) == NULL ||
+				    !spa_pod_is_float(SPA_POD_ARRAY_CHILD(&prop->value)))
 					continue;
-			} else
+				break;
+			default:
 				continue;
-
+			}
 
 			if (c->emitted && c->control.n_values == n_values &&
 			    memcmp(c->control.values, values, sizeof(float) * n_values) == 0)
@@ -1281,6 +1311,9 @@ void pw_stream_destroy(struct pw_stream *stream)
 		free(c);
 	}
 
+	spa_hook_list_clean(&impl->hooks);
+	spa_hook_list_clean(&stream->listener_list);
+
 	spa_hook_remove(&impl->context_listener);
 
 	if (impl->data.context)
@@ -1484,8 +1517,8 @@ pw_stream_connect(struct pw_stream *stream,
 	if (flags & PW_STREAM_FLAG_DONT_RECONNECT)
 		pw_properties_set(stream->properties, PW_KEY_NODE_DONT_RECONNECT, "true");
 
-	impl->warn_mlock = SPA_FLAG_IS_SET(flags, PW_STREAM_FLAG_RT_PROCESS);
-	pw_properties_set(stream->properties, "mem.warn-mlock", impl->warn_mlock ? "true" : "false");
+	impl->process_rt = SPA_FLAG_IS_SET(flags, PW_STREAM_FLAG_RT_PROCESS);
+	pw_properties_set(stream->properties, "mem.warn-mlock", impl->process_rt ? "true" : "false");
 
 	if ((pw_properties_get(stream->properties, PW_KEY_MEDIA_CLASS) == NULL)) {
 		const char *media_type = pw_properties_get(stream->properties, PW_KEY_MEDIA_TYPE);
@@ -1524,6 +1557,7 @@ pw_stream_connect(struct pw_stream *stream,
 	if ((str = pw_properties_get(props, PW_KEY_STREAM_MONITOR)) &&
 	    pw_properties_parse_bool(str)) {
 		pw_properties_set(props, "resample.peaks", "true");
+		pw_properties_set(props, "channelmix.normalize", "true");
 	}
 
 	follower = pw_context_create_node(impl->context, pw_properties_copy(props), 0);
@@ -1559,8 +1593,8 @@ pw_stream_connect(struct pw_stream *stream,
 		pw_properties_free(props);
 		props = NULL;
 	}
-	if (!SPA_FLAG_IS_SET(impl->flags, PW_STREAM_FLAG_INACTIVE))
-		pw_impl_node_set_active(impl->node, true);
+	pw_impl_node_set_active(impl->node,
+			!SPA_FLAG_IS_SET(impl->flags, PW_STREAM_FLAG_INACTIVE));
 
 	pw_log_debug(NAME" %p: export node %p", stream, impl->node);
 	stream->proxy = pw_core_export(stream->core,
@@ -1604,6 +1638,10 @@ int pw_stream_disconnect(struct pw_stream *stream)
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 
 	pw_log_debug(NAME" %p: disconnect", stream);
+
+	if (impl->disconnecting)
+		return 0;
+
 	impl->disconnecting = true;
 
 	if (impl->node)
@@ -1688,18 +1726,18 @@ int pw_stream_set_control(struct pw_stream *stream, uint32_t id, uint32_t n_valu
 
 		if ((c = find_control(stream, id))) {
 			spa_pod_builder_prop(&b, id, 0);
-			switch (c->type) {
+			switch (c->container) {
 			case SPA_TYPE_Float:
-				if (n_values == 1)
-					spa_pod_builder_float(&b, values[0]);
-				else
-					spa_pod_builder_array(&b,
-							sizeof(float), SPA_TYPE_Float,
-							n_values, values);
+				spa_pod_builder_float(&b, values[0]);
 				break;
 			case SPA_TYPE_Bool:
 				spa_pod_builder_bool(&b, values[0] < 0.5 ? false : true);
 				break;
+			case SPA_TYPE_Array:
+				spa_pod_builder_array(&b,
+						sizeof(float), SPA_TYPE_Float,
+						n_values, values);
+				break;
 			default:
 				spa_pod_builder_none(&b);
 				break;
@@ -1838,8 +1876,8 @@ do_flush(struct spa_loop *loop,
 	}
 	while (b);
 
-	impl->time.queued = impl->queued.outcount = impl->dequeued.incount =
-		impl->dequeued.outcount = impl->queued.incount;
+	impl->queued.outcount = impl->dequeued.incount =
+		impl->dequeued.outcount = impl->queued.incount = 0;
 
 	return 0;
 }
@@ -1860,5 +1898,8 @@ int pw_stream_flush(struct pw_stream *stream, bool drain)
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 	pw_loop_invoke(impl->context->data_loop,
 			drain ? do_drain : do_flush, 1, NULL, 0, true, impl);
+	if (!drain)
+		spa_node_send_command(impl->node->node,
+				&SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_Flush));
 	return 0;
 }
